#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the functions of utils.py dealing with logging
"""

import PanACoTA.utils as utils
import logging
import os
import pytest
import shutil

GENEPATH = os.path.join("test", "data", "generated-by-utils-tests")

@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module
    """
    # Init logger to level detail (15)
    # utils.init_logger(LOGFILE_BASE, logging.DEBUG, 'test_utils', verbose=1)
    if os.path.isdir(GENEPATH):
        content = os.listdir(GENEPATH)
        for f in content:
            assert f.startswith(".fuse")
    else:
        os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


def test_class_filter():
    """
    Check that for a class LessThanFilter(warning), info and debug are allowed,
    but warning, error and critical are not.
    """
    a = utils.LessThanFilter(logging.WARNING)
    record = logging.LogRecord("root", logging.DEBUG, "path", 10, "message", "args", "exc_info")
    assert a.filter(record)
    record = logging.LogRecord("root", logging.INFO, "path", 10, "message", "args", "exc_info")
    assert a.filter(record)
    record = logging.LogRecord("root", logging.WARNING, "path", 10, "message", "args", "exc_info")
    assert not a.filter(record)
    record = logging.LogRecord("root", logging.ERROR, "path", 10, "message", "args", "exc_info")
    assert not a.filter(record)
    record = logging.LogRecord("root", logging.CRITICAL, "path", 10, "message", "args", "exc_info")
    assert not a.filter(record)


def test_class_nolevel_filter():
    """
    Check that for a class LessThanFilter(warning), info and debug are allowed,
    but warning, error and critical are not.
    """
    a = utils.NoLevelFilter(logging.WARNING)
    record = logging.LogRecord("root", logging.DEBUG, "path", 10, "message", "args", "exc_info")
    assert a.filter(record)
    record = logging.LogRecord("root", logging.INFO, "path", 10, "message", "args", "exc_info")
    assert a.filter(record)
    record = logging.LogRecord("root", logging.WARNING, "path", 10, "message", "args", "exc_info")
    assert not a.filter(record)
    record = logging.LogRecord("root", logging.ERROR, "path", 10, "message", "args", "exc_info")
    assert a.filter(record)
    record = logging.LogRecord("root", logging.CRITICAL, "path", 10, "message", "args", "exc_info")
    assert a.filter(record)


def test_logger_default(capsys):
    """
    Test that logger is initialized as expected.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.DEBUG
    utils.init_logger(logfile, level, "default")
    logger = logging.getLogger("default")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info error" in err
    assert "info critical" in err
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.details") as logf:
        assert logf.readline().endswith(" :: DETAIL :: info details\n")
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.debug") as logf:
        assert logf.readline().endswith(" :: DEBUG (from default logger) :: info debug\n")
        assert logf.readline().endswith(" :: DETAIL (from default logger) :: info details\n")
        assert logf.readline().endswith(" :: INFO (from default logger) :: info info\n")
        assert logf.readline().endswith(" :: WARNING (from default logger) :: info warning\n")
        assert logf.readline().endswith(" :: ERROR (from default logger) :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL (from default logger) :: info critical\n")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_logger_verbose1(capsys):
    """
    Test that logger is initialized as expected.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.DEBUG
    utils.init_logger(logfile, level, "toto", verbose=1)
    logger = logging.getLogger("toto")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")

    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info warning" in err
    assert "info error" in err
    assert "info critical" in err
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.details") as logf:
        assert logf.readline().endswith(" :: DETAIL :: info details\n")
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.debug") as logf:
        assert logf.readline().endswith(" :: DEBUG (from toto logger) :: info debug\n")
        assert logf.readline().endswith(" :: DETAIL (from toto logger) :: info details\n")
        assert logf.readline().endswith(" :: INFO (from toto logger) :: info info\n")
        assert logf.readline().endswith(" :: WARNING (from toto logger) :: info warning\n")
        assert logf.readline().endswith(" :: ERROR (from toto logger) :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL (from toto logger) :: info critical\n")


def test_logger_verbose2(capsys):
    """
    Test that logger is initialized as expected.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.DEBUG
    utils.init_logger(logfile, level, "toto", verbose=2)
    logger = logging.getLogger("toto")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info debug" in out
    assert "info details" in out
    assert "info info" in out
    assert "info warning" in err
    assert "info error" in err
    assert "info critical" in err
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.details") as logf:
        assert logf.readline().endswith(" :: DETAIL :: info details\n")
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.debug") as logf:
        assert logf.readline().endswith(" :: DEBUG (from toto logger) :: info debug\n")
        assert logf.readline().endswith(" :: DETAIL (from toto logger) :: info details\n")
        assert logf.readline().endswith(" :: INFO (from toto logger) :: info info\n")
        assert logf.readline().endswith(" :: WARNING (from toto logger) :: info warning\n")
        assert logf.readline().endswith(" :: ERROR (from toto logger) :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL (from toto logger) :: info critical\n")


def test_logger_quiet(capsys):
    """
    Test that logger is initialized as expected.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.DEBUG
    utils.init_logger(logfile, level, "quiet", quiet=True)
    logger = logging.getLogger("quiet")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    print(out)
    print(err)
    assert not out
    assert not err
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.details") as logf:
        assert logf.readline().endswith(" :: DETAIL :: info details\n")
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    with open(logfile + ".log.debug") as logf:
        assert logf.readline().endswith(" :: DEBUG (from quiet logger) :: info debug\n")
        assert logf.readline().endswith(" :: DETAIL (from quiet logger) :: info details\n")
        assert logf.readline().endswith(" :: INFO (from quiet logger) :: info info\n")
        assert logf.readline().endswith(" :: WARNING (from quiet logger) :: info warning\n")
        assert logf.readline().endswith(" :: ERROR (from quiet logger) :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL (from quiet logger) :: info critical\n")


def test_logger_info(capsys):
    """
    Test that when logger is initialized with "INFO" level, it does not return DEBUG info.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.INFO
    utils.init_logger(logfile, level, "info")
    logger = logging.getLogger("info")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info error" in err
    assert "info critical" in err
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    assert not os.path.isfile(logfile + ".log.details")
    assert not os.path.isfile(logfile + ".log.debug")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_logger_info_verbose1(capsys):
    """
    Test that when logger is initialized with "INFO" level, it does not return DEBUG info.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.INFO
    utils.init_logger(logfile, level, "info", verbose=1)
    logger = logging.getLogger("info")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info warning" in err
    assert "info error" in err
    assert "info critical" in err
    files = os.listdir(GENEPATH)
    files = [f for f in files if "fuse" not in f]
    assert len(files) == 2
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    assert not os.path.isfile(logfile + ".log.details")
    assert not os.path.isfile(logfile + ".log.debug")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_logger_info_verbose2(capsys):
    """
    Test that when logger is initialized with "INFO" level, it does not return DEBUG info.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.INFO
    utils.init_logger(logfile, level, "info", verbose=2)
    logger = logging.getLogger("info")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info details" in out
    assert "info info" in out
    assert "info warning" in err
    assert "info error" in err
    assert "info critical" in err
    files = os.listdir(GENEPATH)
    files = [f for f in files if "fuse" not in f]
    assert len(files) == 2
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    assert not os.path.isfile(logfile + ".log.details")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_logger_warning(capsys):
    """
    Test that when logger is initialized with "WARNING" level, it does not return
    anything in stdout, as DEBUG and INFO are not returned.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.WARNING
    utils.init_logger(logfile, level, "warn", verbose=1)
    logger = logging.getLogger("warn")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info error" in err
    assert "info critical" in err
    files = os.listdir(GENEPATH)
    files = [f for f in files if "fuse" not in f]
    assert len(files) == 2
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    assert not os.path.isfile(logfile + ".log.details")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_logger_warning_verbose1(capsys):
    """
    Test that when logger is initialized with "WARNING" level, it does not return
    anything in stdout, as DEBUG and INFO are not returned.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.WARNING
    utils.init_logger(logfile, level, "warn", verbose=1)
    logger = logging.getLogger("warn")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info error" in err
    assert "info warning" in err
    assert "info critical" in err
    files = os.listdir(GENEPATH)
    files = [f for f in files if "fuse" not in f]
    assert len(files) == 2
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    assert not os.path.isfile(logfile + ".log.details")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_logger_warning_verbose2(capsys):
    """
    Test that when logger is initialized with "WARNING" level, it does not return
    anything in stdout, as DEBUG and INFO are not returned.
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.WARNING
    utils.init_logger(logfile, level, "warn", verbose=2)
    logger = logging.getLogger("warn")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info details" in out
    assert "info error" in err
    assert "info warning" in err
    assert "info critical" in err
    files = os.listdir(GENEPATH)
    files = [f for f in files if "fuse" not in f]
    assert len(files) == 2
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    assert not os.path.isfile(logfile + ".log.details")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_logger_critical(capsys):
    """
    Test that when logger is initialized with "CRITICAL" level, it only returns
    CRITICAL information.

    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    level = logging.CRITICAL
    utils.init_logger(logfile, level, "crit")
    logger = logging.getLogger("crit")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info error" in err
    assert "info critical" in err
    files = os.listdir(GENEPATH)
    files = [f for f in files if "fuse" not in f]
    assert len(files) == 2
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    assert not os.path.isfile(logfile + ".log.details")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_logger_exists(capsys):
    """
    Test that when the logfiles already exist, it creates new ones with a timestamp added
    """
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    open(logfile + ".log", "w").close()
    open(logfile + ".log.details", "w").close()
    open(logfile + ".log.debug", "w").close()
    open(logfile + ".log.err", "w").close()
    level = logging.DEBUG
    utils.init_logger(logfile, level, "already_exist", verbose=1)
    logger = logging.getLogger("already_exist")
    logger.debug("info debug")
    logger.details("info details")
    logger.info("info info")
    logger.warning("info warning")
    logger.error("info error")
    logger.critical("info critical")
    out, err = capsys.readouterr()
    assert "info info" in out
    assert "info warning" in err
    assert "info error" in err
    assert "info critical" in err
    # Check that initial log files are still empty
    with open(logfile + ".log", "r") as logf:
        assert logf.readlines() == []
    with open(logfile + ".log.debug", "r") as logf:
        assert logf.readlines() == []
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readlines() == []
    with open(logfile + ".log.details", "r") as logf:
        assert logf.readlines() == []
    # Check for new .log file, remove the one which is empty
    import glob
    logs = glob.glob(logfile + "*" + ".log")
    assert len(logs) == 2
    logs.remove(logfile + ".log")
    with open(logs[0], "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info info\n")
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")
    # Same thing for .log.err file
    logs_err = glob.glob(logfile + "*" + ".log.err")
    assert len(logs_err) == 2
    logs_err.remove(logfile + ".log.err")
    with open(logs_err[0], "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: info warning\n")
        assert logf.readline().endswith(" :: ERROR :: info error\n")
        assert logf.readline().endswith(" :: CRITICAL :: info critical\n")


def test_log_listen(capsys):
    """
    Check that when we log to a queue listener, and then handle the logs
    via logger_thread, the logs appear.
    """
    import multiprocessing
    import threading

    # Create Queue, QueueHandler, and log messages to it
    m = multiprocessing.Manager()
    q = m.Queue()
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    logger = logging.getLogger('process')
    logger.debug("debug message")
    logger.log(utils.detail_lvl(), "detail message")
    logger.info("info message")
    logger.warning("warning message")
    logger.error("error message")
    logger.critical("critical message")
    q.put(None)

    # Initialize real logger
    logfile = os.path.join(GENEPATH, "logfile_test.txt")
    utils.init_logger(logfile, 0, '')

    # Listen to QueueHandler and handle messages to stdout/stderr/files
    lp = threading.Thread(target=utils.logger_thread, args=(q,))
    lp.start()
    lp.join()

    out, err = capsys.readouterr()
    assert "info message" in out
    assert "error message" in err
    assert "critical message" in err
    with open(logfile + ".log", "r") as logf:
        assert logf.readline().endswith(" :: INFO :: info message\n")
        assert logf.readline().endswith(" :: WARNING :: warning message\n")
        assert logf.readline().endswith(" :: ERROR :: error message\n")
        assert logf.readline().endswith(" :: CRITICAL :: critical message\n")
    with open(logfile + ".log.details") as logf:
        assert logf.readline().endswith(" :: DETAIL :: detail message\n")
        assert logf.readline().endswith(" :: INFO :: info message\n")
        assert logf.readline().endswith(" :: WARNING :: warning message\n")
        assert logf.readline().endswith(" :: ERROR :: error message\n")
        assert logf.readline().endswith(" :: CRITICAL :: critical message\n")
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readline().endswith(" :: WARNING :: warning message\n")
        assert logf.readline().endswith(" :: ERROR :: error message\n")
        assert logf.readline().endswith(" :: CRITICAL :: critical message\n")


def test_log_no_listen(capsys):
    """
    Check that when we log to a queue listener, but never listen to the queue,
    there is nothing in stderr/stdout/files
    """
    import multiprocessing

    # Create Queue, QueueHandler, and log messages to it
    m = multiprocessing.Manager()
    q = m.Queue()
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    logger = logging.getLogger('process')
    logger.debug("debug message")
    logger.log(utils.detail_lvl(), "detail message")
    logger.info("info message")
    logger.warning("warning message")
    logger.error("error message")
    logger.critical("critical message")
    q.put(None)

    # Initialize real logger
    logfile = os.path.join(GENEPATH, "test_log_listen")
    utils.init_logger(logfile, 0, '')

    assert q.qsize() == 7
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""
    with open(logfile + ".log", "r") as logf:
        assert logf.readlines() == []
    with open(logfile + ".log.details") as logf:
        assert logf.readlines() == []
    with open(logfile + ".log.err", "r") as logf:
        assert logf.readlines() == []


# def test_logger_thread(capsys):
#     """
#     Test that, when we put some fake logs in the Queue given to logger_thread,
#     those logs are given to logger, and printed to err.
#     """
#     import multiprocessing
#     import threading
#     m = multiprocessing.Manager()
#     q = m.Queue()
#     lp = threading.Thread(target=utils.logger_thread, args=(q,))
#     lp.start()
#     q.put(FakeLog("myname", "hello!!"))
#     q.put(FakeLog("other name", "that's me!!!"))
#     q.put(None)
#     lp.join()
#     out, err = capsys.readouterr()
#     print(err)
#     assert out == ""
#     assert "hello!!" in err
#     assert "that's me!!!" in err


# class FakeLog:
#     """
#     Class simulating a logger
#     """

#     def __init__(self, name, text, levelno=100):
#         self.name = name
#         self.text = text
#         self.levelno = levelno
#         self.exc_info = ""
#         self.exc_text = ""
#         self.stack_info = ""

#     def getMessage(self):
#         """
#         returns text of log
#         """
#         return self.text



#!/usr/bin/env python3
# coding: utf-8

"""
Functions used for tests
"""
import importlib.util
import logging
import shlex
import subprocess

import sys

import os


def run_cmd(cmd, error, eof=False, **kwargs):
    """
    Run the given command line. If the return code is not 0, print error message.
    if eof (exit on fail) is True, exit program if error code is not 0.

    Parameters
    ----------
    cmd : str
        command to run
    error : str
        error message to print if error while running command
    eof : bool
        True: exit program if command failed, False: do not exit even if command fails
    kwargs : Object
        Can provide a logger, stdout and/or stderr streams

    Returns
    -------
    subprocess.Popen
        returns object of subprocess call (has attributes returncode, pid, communicate etc.)

    """
    if "logger" not in kwargs:
        logger = logging.getLogger("utils.run_cmd")
    else:
        logger = kwargs["logger"]
    if "stdout" not in kwargs:
        kwargs["stdout"] = None
    if "stderr" not in kwargs:
        kwargs["stderr"] = None
    try:
        call = subprocess.Popen(shlex.split(cmd), stdout=kwargs["stdout"],
                                stderr=kwargs["stderr"])
        call.wait()
        retcode = call.returncode
    except OSError:
        logger.error(error + ": " + "{} does not exist".format(cmd))
        if eof:
            sys.exit(-1)
        else:
            return -1
    if retcode != 0:
        logger.error(error)
        if eof:
            sys.exit(retcode)
    return call


def is_package_installed(pkg_name):
    """
    Check if package is installed or not

    Parameters
    ----------
    pkg_name : str
        name of package to check

    Returns
    -------
    bool
        True if package is installed, False otherwise
    """
    spec = ""
    spec = importlib.util.find_spec(pkg_name)
    return spec is not None


def check_installed(cmd):
    """
    Check if the command 'cmd' is in $PATH and can then be executed

    Parameters
    ----------
    cmd : str
        command to run

    Returns
    -------
    bool
        True if installed, False otherwise
    """
    torun = "which " + cmd
    trying = subprocess.Popen(shlex.split(torun), stdout=subprocess.PIPE)
    out, _ = trying.communicate()
    if trying.returncode == 0:
        if os.path.isfile(out.strip()):
            return True
    return False



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
From the Persistent Genome file, group all persistent proteins per genome, in order to be
able to extract them faster after.

Input:

- Persistent genome (1 line per family, 1st column is family number, others are all members)
- list of all genomes (1 genome per line, only first column is considered)
- output directory
- dname: the name of the dataset, used to name output files

Output:

- for each genome: ``<outdir>/List-<dname>/<dname>-getEntry_gen_<genome>.txt``: list of all
  genes from the 'genome' which correspond to persistent proteins. 2 columns:
  the first one is the protein name, the second is the filename to which it must be extracted
  (corresponding to the family in the persistent genome).
- for each genome: ``<outdir>/List-<dname>/<dname>-getEntry_prt_<genome>.txt``: same as
  previous file but with the list of proteins to extract instead of genes.
- for each family: ``<outdir>/Align-<dname>/<dname>-current.<fam_num>.miss.lst``: list of
  genomes which do not have members in family 'fam_num'.

@author GEM
November 2016
"""

import os
import sys
import logging

logger = logging.getLogger("align.pan_to_pergenome")


def get_per_genome(persgen, list_gen, dname, outdir):
    """
    From persistent genome and list of all genomes, sort persistent proteins by genome

    For each genome, write all persistent proteins to a file, with the family from which they
    are, in order to extract those proteins after.
    For each family, also save the names of genomes which do not have any member. This will
    be used to complete the alignments by stretches of '-'.

    Parameters
    ----------
    persgen : str
        file containing persistent genome
    list_gen : str
        file containing the list of all genomes
    dname : str
        name of the dataset
    outdir : str
        Directory where files must be saved. Will create 2 subfolders: ``Align-<dname>``
        and ``List-<dname>``

    Returns
    -------
    (all_genomes, aldir, listdir, families) : tuple

        * all_genomes : [] list of all genome names
        * aldir : str, path to align directory
        * listdir : str, path to List directory
        * families : str, list of family numbers
    """
    # Define output directories
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    os.makedirs(aldir, exist_ok=True)
    os.makedirs(listdir, exist_ok=True)

    # Get list of all genomes
    all_genomes = get_all_genomes(list_gen)
    logger.info(f"Found {len(all_genomes)} genomes.")

    # Sort proteins by strain
    logger.info("Reading PersGenome and constructing lists of missing genomes in each family.")
    all_prots, fam_genomes, several = proteins_per_strain(persgen)
    # Write output files
    write_getentry_files(all_prots, several, listdir, aldir, dname, all_genomes)
    write_missing_genomes(fam_genomes, several, all_genomes, aldir, dname)
    return all_genomes, aldir, listdir, fam_genomes.keys()


def get_all_genomes(list_gen):
    """
    Read the file containing the genomes list and get all genome names

    Parameters
    ----------
    list_gen : str
        File containing the list of all genomes

    Returns
    -------
    list
        list of all genome names

    """
    all_genomes = []
    with open(list_gen, "r") as lgf:
        for line in lgf:
            if "gembase" not in line:
                genome = line.strip().split()[0]
                all_genomes.append(genome)
    return all_genomes


def proteins_per_strain(persgen):
    """
    From the persistentGenome file, get all persistent proteins, and classify them
    according to the strain from which they are.

    Parameters
    ----------
    persgen : str
        File containing persistent genome

    Returns
    -------
    (all_prots, fam_genomes, several) : tuple

        * all_prots: dict, {strain: {member: fam_num}}
        * fam_genomes: dict, {fam_num: [genomes having a member in fam]}
        * several: dict, {fam_num: [genomes having several members in fam]}
    """
    logger.info("Getting all persistent proteins and classify by strain.")
    all_prots = {}  # {strain: {member: fam_num}}
    fam_genomes = {}  # {fam_num: [genomes having a member in fam]}
    several = {}  # {fam_num: [genomes having several members in fam]}
    with open(persgen, "r") as pgf:
        for line in pgf:
            fam_num = line.split()[0]
            members = line.strip().split()[1:]
            fam_genomes[fam_num] = []
            several[fam_num] = []
            for mem in members:
                # if format is ESCO.1512.00001.i0002_12124, strain is 3 first fields
                # separated by '.'
                if "." in mem and len(mem.split(".")) >= 3:
                    strain = ".".join(mem.split(".")[:3])
                # if format is not like this, it must be something_00001:
                else:
                    strain = "_".join(mem.split("_")[:-1])
                # if strain not already in fam_genomes, add it
                if strain not in fam_genomes[fam_num]:
                    fam_genomes[fam_num].append(strain)
                # If strain already in fam_genomes, it has several members: add it to several
                elif strain not in several[fam_num]:
                    several[fam_num].append(strain)
                if strain not in all_prots:
                    all_prots[strain] = {}
                if mem in all_prots[strain]:
                    logger.warning((" problem: {} already exists, in family {}. Conflict with "
                                    "family {}.").format(mem, all_prots[strain][mem], fam_num))
                all_prots[strain][mem] = fam_num
    return all_prots, fam_genomes, several


def write_getentry_files(all_prots, several, listdir, aldir, dname, all_genomes):
    """
    For each species, write all its persistent proteins into a file, with the
    persistent family in which they are. Those files will be used to extract all
    proteins.

    Parameters
    ----------
    all_prots : dict
        {strain: {member: fam_num}}
    several : dict
        {fam_num: [genomes having several members in family]}
    listdir : str
        directory where lists of proteins per genome must be saved
    aldir : str
         directory where extracted proteins per family must be saved
    dname : str
        name of dataset
    all_genomes : list
        list of all genomes
    """
    for strain, member in all_prots.items():
        write_genome_file(listdir, aldir, dname, strain, member, several)
    error = []
    for strain in all_genomes:
        gegenfile = os.path.join(listdir, dname + "-getEntry_gen_" + strain + ".txt")
        geprtfile = os.path.join(listdir, dname + "-getEntry_prt_" + strain + ".txt")
        if not os.path.isfile(gegenfile) or not os.path.isfile(geprtfile):
            error.append(strain)
    if error:
        for gen in error:
            logger.error(("There is not any protein for genome {} in any family! "
                          "The program will close, please fix this problem to be able to "
                          "run the alignments").format(gen))
        sys.exit(1)


def write_genome_file(listdir, aldir, dname, strain, member, several):
    """
    For a given genome, write all the proteins and genes to extract to its file.
    If one of the 2 files (proteins and genes) already exists, overwrite it.
    If no file exists -> write them.
    If the 2 files exist -> warning saying that we use already existing files.

    Parameters
    ----------
    listdir : str
        path to List directory
    aldir : str
        path to Align directory
    dname : str
        name of dataset
    strain : str
        current genome name
    member : dict
        {member: fam_num}
    several : dict
        {fam_num: [genomes having several members in family]}
    """
    # If the 2 files exist, use them as they are
    gegenfile = os.path.join(listdir, dname + "-getEntry_gen_" + strain + ".txt")
    geprtfile = os.path.join(listdir, dname + "-getEntry_prt_" + strain + ".txt")
    if os.path.isfile(gegenfile) and os.path.isfile(geprtfile):
        logger.warning(f"For genome {strain}, {geprtfile} and {gegenfile} already exist. "
                       "The program will use them "
                       "to extract proteins and genes. If you prefer to rewrite them, use "
                       "option -F (or --force).")
        return

    # If at least one of the 2 files already exists, overwrite both files
    with open(gegenfile, "w") as gegf, open(geprtfile, "w") as gepf:
        for mem, fam in member.items():
            if strain not in several[fam]:
                genfile = os.path.join(aldir, dname + "-current." + fam + ".gen")
                gegf.write(mem + " " + genfile + "\n")
                prtfile = os.path.join(aldir, dname + "-current." + fam + ".prt")
                gepf.write(mem + " " + prtfile + "\n")


def write_missing_genomes(fam_genomes, several, all_genomes, aldir, dname):
    """
    For each family, write the names of all genomes which do not have any member in
    the family.

    Parameters
    ----------
    fam_genomes : dict
        {fam_num: [genomes having a member in fam]}
    several : dict
        {fam_num: [genomes having several members in family]}
    all_genomes : list
        list of all genomes
    aldir : str
        directory where the lists of missing genomes per family must be saved
    dname : str
        name of dataset
    """
    for fam, genomes in fam_genomes.items():
        # File where missing genomes will be written
        missfile = os.path.join(aldir, f"{dname}-current.{fam}.miss.lst")
        with open(missfile, "w") as mff:
            # missing = missing or several members:
            # miss: all genomes - genomes in the family
            # several: add to 'miss' genomes with several members in the family
            missing = (set(all_genomes) - set(genomes)).union(set(several[fam]))
            if missing:
                mff.write("\n".join(missing) + "\n")



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Subcommand to prepare a dataset:

- Download all genomes of a given species from refseq
- Filter them with L90 and number of contigs thresholds
- Remove too close/far genomes using Mash


@author Amandine PERRIN
August 2019
"""
import os
import sys
import logging
from termcolor import colored

from PanACoTA import __version__ as version
from PanACoTA import utils
from PanACoTA.prepare_module import download_genomes_func as dgf
from PanACoTA.prepare_module import filter_genomes as fg


def main_from_parse(arguments):
    """
    Call main function from the arguments given by parser

    Parameters
    ----------
    arguments : argparse.Namespace
        result of argparse parsing of all arguments in command line

    """
    cmd = "PanACoTA " + ' '.join(arguments.argv)
    main(cmd, arguments.ncbi_species_name, arguments.ncbi_species_taxid, arguments.ncbi_taxid, arguments.levels,
         arguments.ncbi_section, arguments.outdir, arguments.tmp_dir, arguments.parallel, arguments.norefseq,
         arguments.db_dir, arguments.only_mash,
         arguments.info_file, arguments.l90, arguments.nbcont, arguments.cutn, arguments.min_dist,
         arguments.max_dist, arguments.verbose, arguments.quiet)


def main(cmd, ncbi_species_name, ncbi_species_taxid, ncbi_taxid, levels, ncbi_section,
         outdir, tmp_dir, threads, norefseq, db_dir,
         only_mash, info_file, l90, nbcont, cutn, min_dist, max_dist, verbose, quiet):
    """
    Main method, constructing the draft dataset for the given species

    verbosity:
    - defaut 0 : stdout contains INFO, stderr contains ERROR, .log contains INFO and more, .log.err contains warning and more
    - 1: same as 0 + WARNING in stderr
    - 2: same as 1 + DETAILS in stdout + DETAILS in .log.details
    - >=15: same as 2 + Add DEBUG in stdout + create .log.debug with everything from info to debug


    Parameters
    ----------
    cmd : str
        command line used to launch this program
    ncbi_species_name : str
        name of species to download, as given by NCBI
    ncbi_species_taxid : int
        species taxid given in NCBI
    ncbi_taxid : int
        NCBI taxid of strain
    levels: str
        Level of assembly to download. Choice between 'all', 'complete', 'chromosome',
        'scaffold', 'contig'. Default is 'all'
    outdir : str
        path to output directory (where created database will be saved).
    tmp_dir : str
        Path to directory where tmp files are saved (sequences split at each row of 5 'N')
    threads : int
        max number of threads to use
    norefseq : bool
        True if user does not want to download again the database
    db_dir : str
        Name of the folder where already downloaded fasta files are saved.
    only_mash : bool
        True if user user already has the database and quality of each genome (L90, #contigs etc.)
    info_file : str
        File containing information on QC if it was already ran before (columns to_annotate,
        gsize, nb_conts and L90).
    l90 : int
        Max L90 allowed to keep a genome
    nbcont : int
        Max number of contigs allowed to keep a genome
    cutn : int
        cut at each when there are 'cutn' N in a row. Don't cut if equal to 0
    min_dist : int
        lower limit of distance between 2 genomes to keep them
    max_dist : int
        upper limit of distance between 2 genomes to keep them (default is 0.06)
    verbose : int
        verbosity:
        - defaut 0 : stdout contains INFO, stderr contains ERROR, .log contains INFO and more,
          .log.err contains warning and more
        - 1: same as 0 + WARNING in stderr
        - 2: same as 1 + DETAILS in stdout + DETAILS in .log.details
        - >=15: same as 2 + Add DEBUG in stdout + create .log.debug with everything
          from info to debug
    quiet : bool
        True if nothing must be sent to stdout/stderr, False otherwise
    """

    # get species name in NCBI format
    # -> will be used to name output directory
    # -> will be used to download summary file if given species corresponds to NCBI name
    if ncbi_species_name:
        species_linked = "_".join(ncbi_species_name.split())
        species_linked = "_".join(species_linked.split("/"))

    # if species name not given by user, use species taxID (if given) to name output directory
    elif ncbi_species_taxid:
        species_linked = str(ncbi_species_taxid)
    # if species name not species taxid by user, use taxID (if given) to name output directory
    elif ncbi_taxid:
        species_linked = str(ncbi_taxid)
    # if neither speName, speID nor taxID given (--norefseq, mashonly), name is NA
    else:
        species_linked = "NA"
    # Default outdir is species name if given, or species taxID
    if not outdir:
        outdir = species_linked
    # Default tmp_dir is outdir/tmp_files
    if not tmp_dir:
        tmp_dir = os.path.join(outdir, "tmp_files")
    # directory that will be created by ncbi_genome_download
    ncbidir = os.path.join(outdir, ncbi_section, "bacteria")
    os.makedirs(outdir, exist_ok=True)
    os.makedirs(tmp_dir, exist_ok=True)

    # Initialize logger
    # set level of logger: level is the minimum level that will be considered.
    if verbose <= 1:
        level = logging.INFO
    # for verbose = 2, ignore only debug
    if verbose >= 2 and verbose < 15:
        level = utils.detail_lvl() # int corresponding to detail level
    # for verbose >= 15, write everything
    if verbose >= 15:
        level = logging.DEBUG
    logfile_base = os.path.join(outdir, "PanACoTA_prepare_{}").format(species_linked)
    logfile_base, logger = utils.init_logger(logfile_base, level, 'prepare', log_details=True,
                                             verbose=verbose, quiet=quiet)

    # Message on what will be done (cmd, cores used)
    logger.info(f'PanACoTA version {version}')
    logger.info("Command used\n \t > " + cmd)
    message = f"'PanACoTA prepare' will run on {threads} "
    message += f"cores" if threads>1 else "core"
    logger.info(message)

    # Start prepare step
    # Run more than only mash filter (!only_mash):
    # - start from QC and mash (norefseq)
    # - start from genome download (!norefseq))
    if not only_mash:
        # Not only mash, so a new info file will be created. If the user still gave an info
        # file (he will be warned that it will be ignored), rename it with '.bak'
        # to avoid erasing it
        if info_file and os.path.isfile(info_file):
            os.rename(info_file, info_file + ".back")

        # 'norefseq = True" : Do not download genomes, just do QC and mash filter on given genomes
        # -> if not, error and exit
        if norefseq:
            logger.warning(f'You asked to skip {ncbi_section} downloads.')

            # -> if db_dir given, watch for sequences there. If does not exist, error and exit
            # (user gave a directory (even if it does not exist), so we won't look for
            # the sequences in other folders)
            if db_dir:
                if not os.path.exists(db_dir):
                    logger.error(f"Database folder {db_dir} supposed to contain fasta "
                                 "sequences does not "
                                 "exist. Please give a valid folder, or leave the default "
                                 "directory (no '-d' option).")
                    sys.exit(1)
            # -> If user did not give db_dir, genomes could be in
            # outdir/Database_init/<genome_name>.fna
            else:
                db_dir = os.path.join(outdir, "Database_init")
                # If it does not exist, check if default compressed files folder exists.
                if not os.path.exists(db_dir):
                    logger.warning(f"Database folder {db_dir} supposed to contain fasta "
                                   "sequences does not "
                                   "exist. We will check if the download folder (with compressed "
                                   "sequences) exists.")
                    # -> if not in database_init, genomes must be in
                    # outdir/refeq/bacteria/<genome_name>.fna.gz. In that case,
                    # uncompress and add them to Database_init
                    if not os.path.exists(ncbidir):
                        logger.error(f"Folder {ncbidir} does not exist. You do not have any "
                                     "genome to analyse. Possible reasons:\n"
                                     "- if you want to rerun analysis in the same folder as "
                                     "sequences were downloaded (my_outdir/Database_init or "
                                     f"my_outdir/{ncbi_section}), make sure you have '-o my_outdir' "
                                     "option\n"
                                     "- if you want to rerun analysis and save them in a new "
                                     "output folder called 'new_outdir', make sure you have "
                                     "'-o new_outdir' option, "
                                     "and you specified where the uncompressed sequences to "
                                     "use are ('-d sequence_database_path'). ")
                        sys.exit(1)
                    # add genomes from refseq/bacteria folder to Database_init
                    nb_gen, _ = dgf.to_database(outdir, ncbi_section)
        # No sequence: Do all steps -> download, QC, mash filter
        else:
            # Download all genomes of the given taxID
            db_dir, nb_gen = dgf.download_from_ncbi(species_linked, ncbi_section, ncbi_species_name, ncbi_species_taxid,
                                                      ncbi_taxid, levels, outdir, threads)
            logger.info(f"{nb_gen} {ncbi_section} genome(s) downloaded")

        # Now that genomes are downloaded and uncompressed, check their quality to remove bad ones
        genomes = fg.check_quality(species_linked, db_dir, tmp_dir, l90, nbcont, cutn)

    # Do only mash filter. Genomes must be already downloaded, and there must be a file with
    # all information on these genomes (L90 etc.)
    else:
        logger.warning('You asked to run only mash steps.')
        if not os.path.exists(info_file):  # info-file missing -> error and exit
            logger.error(f"Your info file {info_file} does not exist. Please provide the  "
                          "right name/path, or remove the '--mash-only option to rerun "
                          "quality control.")
            sys.exit(1)
        logger.info(("You want to run only mash steps. Getting information "
                     "from {}").format(info_file))
        genomes = utils.read_genomes_info(info_file, species_linked, )

    # Run Mash
    # genomes : {genome_file: [genome_name, orig_name, path_to_seq_to_annotate, size, nbcont, l90]}
    # sorted_genome : [genome_file] ordered by L90/nbcont (keys of genomes)
    sorted_genomes = fg.sort_genomes_minhash(genomes, l90, nbcont)

    # Write discarded genomes to a file -> orig_name, to_annotate, gsize, nb_conts, L90
    discQC = f"by-L90_nbcont-{species_linked}.txt"
    utils.write_genomes_info(genomes, sorted_genomes, discQC, outdir)

    # Remove genomes not corresponding to mash filters
    removed = fg.iterative_mash(sorted_genomes, genomes, outdir, species_linked,
                                min_dist, max_dist, threads, quiet)
    # Write list of genomes kept, and list of genomes discarded by mash step
    info_file = fg.write_outputfiles(genomes, sorted_genomes, removed, outdir, species_linked,
                                     min_dist, max_dist)
    logger.info("End")
    return info_file


def build_parser(parser):
    """
    Method to create a parser for command-line options

    Parameters
    ----------
    parser : argparse.ArgumentParser
        parser to configure in order to extract command-line arguments
    """
    import argparse
    from PanACoTA import utils_argparse

    general = parser.add_argument_group('General arguments')
    general.add_argument("-T", dest="ncbi_species_taxid", default="",
                          help=("Species taxid to download, corresponding to the "
                                "'species taxid' provided by the NCBI. "
                                "This will download all sequences of this species and all its sub-species."
                                "A comma-separated list of species taxids can also be provided. "
                                "(Ex: -T 573 for Klebsiella pneumoniae)")
                         )
    general.add_argument("-t", dest="ncbi_taxid", default="",
                          help=("Taxid to download. "
                                "This can be the taxid of a sub-species, or of a specific strain. "
                                "A taxid of a subspecies will download all strains in this subspecies "
                                "EXCEPT the ones which have a specific taxid."
                                "A comma-separated list of taxids can also be provided."
                                "Ex: '-t 72407' will download all 'general' K. pneumoniae subsp. pneumoniae strains, "
                                "and '-t 1123862' will download the strain K. pneumoniae subsp. pneumoniae Kp13 "
                                "(not included in -t 72407, as it is a strain of the subspecies with a specific taxid).")
                         )
    general.add_argument("-g", dest="ncbi_species_name", default="",
                          help=("Species to download, corresponding to the "
                                "'organism name' provided by the NCBI. Give name between "
                                "quotes (for example \"escherichia coli\")")
                        )
    general.add_argument("-s", dest="ncbi_section", default="refseq", choices = ["refseq", "genbank"],
                          help=("NCBI section to download: all genbank, or only refseq (default)")
                        )
    general.add_argument("-l", "--assembly_level", dest="levels", default="",
                          help=("Assembly levels of genomes to download (default: all). "
                                "Possible levels are: 'all', 'complete', 'chromosome', "
                                "'scaffold', 'contig'."
                                "You can also provide a comma-separated list of assembly "
                                "levels. For ex: 'complete,chromosome'")
                          )
    general.add_argument("-o", dest="outdir",
                          help=("Give the path to the directory where you want to save the "
                               "downloaded database. In the given directory, it will create "
                               "a folder 'Database_init' containing all fasta "
                               "files that will be sent to the control procedure, as well as "
                               "a folder 'refseq' with all original compressed files "
                               "downloaded from refseq. By default, this output dir name is the "
                               "ncbi_species name if given, or ncbi_species_taxid or ncbi_taxid otherwise.")
                          )
    general.add_argument("--tmp", dest="tmp_dir",
                          help=("Specify where the temporary files (sequence split by rows "
                                "of 'N', sequence with new contig names etc.) must be saved. "
                                "By default, it will be saved in your "
                                "out_dir/tmp_files.")
                          )
    general.add_argument("--cutn", dest="cutn", type=utils_argparse.positive_int, default=5,
                          help=("By default, each genome will be cut into new contigs when "
                                "at least 5 'N' in a row are found in its sequence. "
                                "If you don't want to "
                                "cut genomes into new contigs when there are rows of 'N', "
                                "put 0 to this option. If you want to cut from a different number "
                                "of 'N' in a row, put this value to this option.")
                          )
    general.add_argument("--l90", dest="l90", type=int, default=100,
                          help="Maximum value of L90 allowed to keep a genome. Default is 100.")
    general.add_argument("--nbcont", dest="nbcont", type=utils_argparse.cont_num, default=999,
                          help=("Maximum number of contigs allowed to keep a genome. "
                                "Default is 999."))
    general.add_argument("--min_dist", dest="min_dist", default=1e-4,
                        type=utils_argparse.mash_dist,
                        help="By default, genomes whose distance to the reference is not "
                             "between 1e-4 and 0.06 are discarded. You can specify your own "
                             "lower limit (instead of 1e-4) with this option.")
    general.add_argument("--max_dist", dest="max_dist", default=0.06,
                         type=utils_argparse.mash_dist,
                         help="By default, genomes whose distance to the reference is not "
                              "between 1e-4 and 0.06 are discarded. You can specify your own "
                              "lower limit (instead of 0.06) with this option.")
    general.add_argument("-p", "--threads", dest="parallel", type=utils_argparse.thread_num,
                         default=1, help=("Run 'N' downloads in parallel (default=1). Put 0 if "
                                "you want to use all cores of your computer."))

    optional = parser.add_argument_group('Alternatives')
    optional.add_argument("--norefseq", dest="norefseq", action="store_true",
                          help=("If you already downloaded refseq genomes and do not want to "
                                "check them, add this option to directly go to the next steps:"
                                "quality control (L90, number of contigs...) and mash filter. "
                                "Don't forget to specify the db_dir (-d option) where you "
                                "already have your genome sequences."))
    optional.add_argument("-d", dest="db_dir",
                          help=("If your already downloaded sequences are not in the default "
                                "directory (outdir/Database_init), you can specify here the "
                                "path to those fasta files."))
    optional.add_argument('-M', '--only-mash', dest="only_mash", action="store_true",
                          help=("Add this option if you already downloaded complete and refseq "
                                "genomes, and ran quality control (you have, a file "
                                "containing all genome names, as well as their genome size, "
                                "number of contigs and L90 values). "
                                "It will then get information on genomes quality from this "
                                "file, and run mash steps."))
    optional.add_argument("--info", dest="info_file",
                          help=("If you already ran the quality control, specify from which "
                                "file PanACoTA can read this information, in order to proceed "
                                "to the mash step. This file must contain at "
                                "least 4 columns, tab separated, with the following headers: "
                                "'to_annotate', 'gsize', 'nb_conts', 'L90'. Any other column "
                                "will be ignored."))

    helper = parser.add_argument_group('Others')
    helper.add_argument("-v", "--verbose", dest="verbose", action="count", default=0,
                        help="Increase verbosity in stdout/stderr and log files.")
    helper.add_argument("-q", "--quiet", dest="quiet", action="store_true", default=False,
                        help=("Do not display anything to stdout/stderr. log files will "
                              "still be created."))
    helper.add_argument("-h", "--help", dest="help", action="help",
                        help="show this help message and exit")


def parse(parser, argu):
    """
    Parse arguments given to parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        the parser used
    argu : [str]
        command-line given by user, to parse using parser

    Returns
    -------
    argparse.Namespace
        Parsed arguments
    """
    import argparse

    args = parser.parse_args(argu)
    return check_args(parser, args)


def check_args(parser, args):
    """
    Check that arguments given to parser are as expected.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser used to parse command-line
    args : argparse.Namespace
        Parsed arguments

    Returns
    -------
    argparse.Namespace or None
        The arguments parsed, updated according to some rules. Exit program
        with error message if error occurs with arguments given.

    """
    # Message if user kept default thresholds for L90 and nbcont. Just to warn him, to be sure
    # it was on purpose
    def thresholds_message(l90, nbcont):
        return ("  !! Your genomes will be filtered, and only the ones with 'L90' <= {} "
                "and 'number of contigs' < {} will be kept. If you want to change those "
                "thresholds, use '--l90' and '--nbcont' "
                "options.".format(args.l90, args.nbcont))

    #  ERRORS

    # We don't want to run only mash, nor only quality control, but don't give a NCBI taxID.
    # -> Give at least 1!
    if (not args.only_mash and not args.norefseq and
        not args.ncbi_species_taxid and not args.ncbi_species_name and not args.ncbi_taxid):
        parser.error("As you did not put the '--norefseq' nor the '-M' option, it means that "
                     "you want to download refseq (or genbank) genomes. But you did not provide any "
                     "information, so PanACoTA cannot guess which species you want to download. "
                     "Specify NCBI_taxid (-t), and/or NCBI species taxid (-T) and/or NCBI_species (-g) to download, or add one of "
                     "the 2 options (--norefseq or -M) if you want to skip the 'download step'.")

    # If norefseq, give output directory
    #  - folder containing Database_init, with all sequences
    #  - or new folder where you want to put the new results
    if args.norefseq and not args.outdir:
        parser.error("You must provide an output directory, where your results will be saved.")

    # If user wants only mash steps, check that he gave info file, and outdir
    if args.only_mash and not args.info_file:
        parser.error("If you want to run only Mash filtering steps, please give the "
                     "info file with the required information (see '--info' option)")
    if args.only_mash and not args.outdir:
        parser.error("If you want to run only Mash filtering steps, please give the "
                     "output directory where you want to save your results (see '-o' option)")

    # Cannot be verbose and quiet at the same time
    if int(args.verbose) > 0 and args.quiet:
        parser.error("Choose between a verbose output (-v) or a quiet output (-q)."
                     " You cannot have both.")

    # min_dist must be higher than max_dist
    if float(args.min_dist) >= float(args.max_dist):
        parser.error(f"min_dist ({args.min_dist}) cannot be higher "
                     f"than max_dist ({args.max_dist})")

    # Check that levels, if given, are among possible ones
    possible = ["all", "complete", "chromosome", "scaffold", "contig"]
    if args.levels:
        for level in args.levels.split(","):
            if level not in possible:
                parser.error("Please choose between available assembly levels: 'all', 'complete', "
                             "'chromosome', 'scaffold', 'contig'. If several levels, provide a "
                             f"comma-separated list. Invalid value: '{args.levels}'")

    # WARNINGS
    # User did not specify a species name
    if not args.ncbi_species_name and not args.outdir:
        if args.ncbi_species_taxid:
            print(colored("WARNING: you did not provide a species name ('-g species' option) "
                          "nor an output directory ('-o outdir'). "
                          "All files will be downloaded in a folder called with the NCBI species "
                          f"taxid {args.ncbi_species_taxid} instead of the species name.", "yellow"))
        else:
            print(colored("WARNING: you did not provide a species name ('-g species' option) "
                          "nor a species taxid ('-T spetaxid') nor an output directory ('-o outdir'). "
                          "All files will be downloaded in a folder called with the NCBI "
                          f"taxid {args.ncbi_taxid}.", "yellow"))
    # If user wants to cut genomes, warn him to check that it is on purpose (because default is cut at each 5'N')
    if args.cutn == 5:
        message = ("  !! Your genomes will be split when sequence contains at "
                   "least {}'N' in a row. If you want to change this threshold, use "
                   "'--cutn n' option (n=0 if you do not want to cut)").format(args.cutn)
        print(colored(message, "yellow"))

    # Warn user about selection of genomes thresholds
    if args.l90 == 100 or args.nbcont == 999:
        print(colored(thresholds_message(args.l90, args.nbcont), "yellow"))

    # Warn if user gave info file, but does not ask to run only Mash -> info file will be ignored
    if (args.info_file and not args.only_mash) or (args.info_file and not args.norefseq):
        message = ("  !! You gave an info file (--info option), but did not ask to run only Mash "
                   "step (-M option). Your info file will be ignored (and renamed with '.back' "
                   "at the end), and another one will "
                   "be created with the new calculated values.")
        print(colored(message, "yellow"))
    return args



if __name__ == '__main__':
    import argparse
    from textwrap import dedent
    header = '''
     ___                 _____  ___         _____  _____
    (  _`\              (  _  )(  _`\      (_   _)(  _  )
    | |_) )  _ _   ___  | (_) || ( (_)   _   | |  | (_) |
    | ,__/'/'_` )/' _ `\|  _  || |  _  /'_`\ | |  |  _  |
    | |   ( (_| || ( ) || | | || (_( )( (_) )| |  | | | |
    (_)   `\__,_)(_) (_)(_) (_)(____/'`\___/'(_)  (_) (_)


       Large scale comparative genomics tools

     -------------------------------------------
     '''
    my_parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                        description=dedent(header), add_help=False)
    build_parser(my_parser)
    OPTIONS = parse(my_parser, sys.argv[1:])
    main_from_parse(OPTIONS)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
For a given family:

- align its proteins using mafft
- back translate to nucleotides
- add "-" of same size as alignment for genomes not having members in the family

@author: GEM, Institut Pasteur
March 2017
"""

import os
import sys
import logging
import multiprocessing
import progressbar
import threading

from PanACoTA import utils

main_logger = logging.getLogger("align.alignment")


def align_all_families(prefix, all_fams, ngenomes, dname, quiet, threads):
    """
    For each family:

    - align all its proteins with mafft
    - back-translate to nucleotides
    - add missing genomes

    Parameters
    ----------
    prefix :  str
        path to ``aldir/<name of dataset>`` (used to get extraction, alignment and btr files
        easily)
    all_fams : []
        list of all family numbers
    ngenomes : int
        total number of genomes in dataset
    dname : str
        name of dataset (used to name concat and grouped files, as well as tree folder)
    quiet : bool
        True if nothing must be written in stdout/stderr, False otherwise
    threads : int
        max number of threads that can be used by mafft

    Returns
    -------
    bool
        True if everything went well, False if there was a problem in at least 1 family.
    """
    main_logger.info(("Starting alignment of all families: protein alignment, "
                      "back-translation to nucleotides, and add missing genomes in the family"))
    nbfam = len(all_fams)
    bar = None
    if not quiet:
        # Create progressbar
        widgets = ['Alignment: ', progressbar.Bar(marker='█', left='', right='', fill=' '),
                   ' ', progressbar.Counter(), "/{}".format(nbfam), ' (',
                   progressbar.Percentage(), ') - ', progressbar.Timer(), ' - '
                   ]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=nbfam,
                                      term_width=79).start()
    final = []
    if threads == 1:
        update_bar = 1
        for num_fam in all_fams:
            f = handle_family_1thread((prefix, num_fam, ngenomes))
            final.append(f)
            bar.update(update_bar)
            update_bar+=1

    else:
        pool = multiprocessing.Pool(threads)

        # Create a Queue to put logs from processes, and handle them after from a single thread
        m = multiprocessing.Manager()
        q = m.Queue()
        # arguments : (gpath, cores_prokka, name, force, nbcont, q) for each genome
        arguments = [(prefix, num_fam, ngenomes, q) for num_fam in all_fams]
        try:
            final = pool.map_async(handle_family, arguments, chunksize=1)
            pool.close()
            # Listen for logs in processes
            lp = threading.Thread(target=utils.logger_thread, args=(q,))
            lp.start()
            if not quiet:
                while True:
                    if final.ready():
                        break
                    remaining = final._number_left
                    bar.update(nbfam - remaining)
                bar.finish()
            pool.join()
            q.put(None)
            lp.join()
            final = final.get()
        # If an error occurs (or user kills with keybord), terminate pool and exit
        except Exception as excp:  # pragma: no cover
            pool.terminate()
            main_logger.error(excp)
            sys.exit(1)
        # We re-aligned at least one family -> remove concatenated files and groupby
    if set(final) != {"OK"}:
        aldir = os.path.split(prefix)[0]
        concat = os.path.join(aldir, f"{dname}-complete.cat.aln")
        outdir = os.path.split(aldir)[0]
        treedir = os.path.join(outdir, "Phylo-" + dname)
        grpfile = os.path.join(treedir, dname + ".grp.aln")
        utils.remove(concat)
        utils.remove(grpfile)
    return False not in final


def handle_family_1thread(args):
    """
    For the given family:

    - align its proteins with mafft
    - back-translate to nucleotides
    - add missing genomes

    Parameters
    ----------
    args : ()
         (prefix, num_fam, ngenomes, q) with:

         - prefix: path to ``aldir/<name of dataset>``
         - num_fam: the current family number
         - ngenomes: the total number of genomes in dataset

    Returns
    -------
    bool
        - "OK" if the files were not re-created, and have the expected format. This is used by
          ``align_all_families`` function, to know if something was regenerated, or if everything
          already existed with the expected format. If something was redone and concat/group files
          exist, it removes them.
        - False if any problem (extractions, alignment, btr, add missing genomes...)
        - True if just generated all files, and everything is ok
    """
    prefix, num_fam, ngenomes = args
    logger = logging.getLogger('align.align_family')
    # Get file names
    prt_file = "{}-current.{}.prt".format(prefix, num_fam)
    gen_file = "{}-current.{}.gen".format(prefix, num_fam)
    miss_file = "{}-current.{}.miss.lst".format(prefix, num_fam)
    mafft_file = "{}-mafft-align.{}.aln".format(prefix, num_fam)
    btr_file = "{}-mafft-prt2nuc.{}.aln".format(prefix, num_fam)
    status1 = family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file,
                               num_fam, ngenomes, logger)
    # If it returned true, Add missing genomes
    if status1:
        return add_missing_genomes(btr_file, miss_file, num_fam, ngenomes, status1, logger)
    return False


def handle_family(args):
    """
    For the given family:

    - align its proteins with mafft
    - back-translate to nucleotides
    - add missing genomes

    Parameters
    ----------
    args : ()
         (prefix, num_fam, ngenomes, q) with:

         - prefix: path to ``aldir/<name of dataset>``
         - num_fam: the current family number
         - ngenomes: the total number of genomes in dataset
         - q: a queue, which will be used by logger to put logs while in other process

    Returns
    -------
    bool
        - "OK" if the files were not re-created, and have the expected format. This is used by
          ``align_all_families`` function, to know if something was regenerated, or if everything
          already existed with the expected format. If something was redone and concat/group files
          exist, it removes them.
        - False if any problem (extractions, alignment, btr, add missing genomes...)
        - True if just generated all files, and everything is ok
    """
    prefix, num_fam, ngenomes, q = args
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    logger = logging.getLogger('align.align_family')
    return handle_family_1thread((prefix, num_fam, ngenomes))


def add_missing_genomes(btr_file, miss_file, num_fam, ngenomes, status1, logger):
    """
    Once all family proteins are aligned, and back-translated to nucleotides,
    add missing genomes for the family to the alignment with '-'.

    Parameters
    ----------
    btr_file : str
        path to file containing protein alignments back-translated to nucleotides
    miss_file : str
        path to file containing the list of missing genomes in this family
    num_fam : int
        family number
    ngenomes : int
        total number of genomes in dataset
    status1 : bool or str
        - "OK" if we did not redo the alignments as they already were as expected. In that case,
          if missing genomes are already present, just add a warning message saying that we
          used the already existing btr file.
        - True if we just did the alignments and backtranslate. So no warning message needed.
        - False if problem with extraction, alignment or backtranslation (will never happen as
          this function is not called if status1 == False)
    logger : logging.Logger
        the logger, having a queue Handler, to give logs to the main logger

    Returns
    -------
    bool or str
        - "OK" if btr file was not recreated, and already has the right number of sequences,
          and all with the same length.
        - False if problem in btr file alignment, so missing genomes not added
        - True if alignment + adding missing genomes is ok. Can happen if there is no missing
          genome for this family (in that case, btr generated already has the right number of
          sequences), or if we just added the missing genomes.

    """
    # btr_file should always exist.
    # Sometimes it comes from previous step ('missing genomes' are missing)
    # Sometimes it comes from a previous run (all genomes should be here)
    status = check_add_missing(btr_file, num_fam, ngenomes, logger, prev=True)
    # If btr_file has the correct number of sequences, all the same length, return True
    if status is True:
        if status1 == "OK":
            logger.warning(("Alignment already done for family {}. The program will use "
                            "it for next steps").format(num_fam))
            return "OK"
        else:
            return True
    # If btr_files has problem in alignment (not all sequences with same size)
    elif status is False:
        return False
    # All sequences have same length but some genomes are missing -> Add missing genomes
    logger.log(utils.detail_lvl(), "Adding missing genomes for family {}".format(num_fam))
    len_aln = status
    with open(miss_file, "r") as missf, open(btr_file, "a") as btrf:
        for genome in missf:
            genome = genome.strip()
            toadd = ">" + genome + "\n" + "-" * len_aln + "\n"
            btrf.write(toadd)
    ret = check_add_missing(btr_file, num_fam, ngenomes, logger)
    # Return True only if all genomes found with all same length (check_add_missing = True,
    # and not check_add_missing = len_alignment)
    if ret is True:
        return True
    else:
        return False


def check_add_missing(btr_file, num_fam, ngenomes, logger, prev=False):
    """
    Check back-translated alignment while missing genomes have been added

    Parameters
    ----------
    btr_file : str
        path to file containing back-translated alignment
    num_fam : int
        current family number
    ngenomes : int
        total number of genomes in dataset
    logger : logging.Logger
        logger with queueHandler to give logs to main logger
    prev : bool
        True if we are checking alignments before adding missing genomes (in case it comes
        from a previous run and is already done, or in case there are no missing genomes for
        this family, so nothing to do to btr file). In that case, do not write error message
        if the number of sequences does not correspond to the total number of genomes.
        False if we just added missing genomes. In that case, nb sequences should be equal to
        the total number of genomes. If not, write error message.

    Returns
    -------
    bool or int
        - True if right number of sequences in btr file and all the same length (everything is ok)
        - False if problem in alignment (sequences do not all have the same size)
        - alignment length if sequences aligned all have the same length, but missing
          genomes are not added yet (so, will have to add lines with this number of '-')
    """
    res = check_lens(btr_file, num_fam, logger)
    # If all sequences have the same length, get number of sequences
    if res:
        len_aln, nb = res
    # otherwise, return False: problem while aligning or back-translating
    else:
        return False
    if nb != ngenomes:
        if not prev:
            logger.error(("ERROR: family {} contains {} genomes in total instead of the {} "
                          "genomes in input.\n").format(num_fam, nb, ngenomes))
        return len_aln
    return True


def family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file,
                     num_fam, ngenomes, logger):
    """
    From a given family, align all its proteins with mafft, back-translate
    to nucleotides, and add missing genomes in this family.

    Parameters
    ----------
    prt_file : str
        path to file containing proteins extracted
    gen_file : str
        path to file containing genes extracted
    miss_file : str
        path to file containing list of genomes missing
    mafft_file : str
        path to file which will contain the protein alignment
    btr_file : str
        path to file which will contain the nucleotide alignment back-translated from protein
        alignment
    num_fam : int
        current family number
    ngenomes : int
        total number of genomes in dataset
    logger : logging.Logger
        logger with queueHandler to give logs to main logger

    Returns
    -------
    bool or str
        - False if problem with extractions or with alignment or with backtranslation
        - 'nb_seqs' = number of sequences aligned if everything went well (extractions and
          alignment ok, btr created without problem)
        - "OK" if extractions and alignments went well, and btr already exists and is ok
    """
    # Check number of proteins extracted
    nbfprt = check_extractions(num_fam, miss_file, prt_file, gen_file, ngenomes, logger)
    nbfal = None
    # If problem with extractions, remove mafft and btr files if they exist, so that they will
    # be regenerated
    if not nbfprt:
        utils.remove(mafft_file)
        utils.remove(btr_file)
        return False
    # If mafft file already exists, check the number of proteins aligned corresponds to number of
    #  proteins extracted. If not, remove mafft and btr files.
    if os.path.isfile(mafft_file):
        message = ("fam {}: Will redo alignment, because found a different number of proteins "
                   "extracted in {} ({}) and proteins aligned in "
                   "existing {}").format(num_fam, prt_file, nbfprt, mafft_file)
        nbfal = check_nb_seqs(mafft_file, nbfprt, logger, message)
        if not nbfal:
            os.remove(mafft_file)
            utils.remove(btr_file)
    # If mafft file does not exist (removed because problem in its alignment, or just not generated
    # yet), remove btr (will be regenerated), and do alignment with mafft
    if not os.path.isfile(mafft_file):
        utils.remove(btr_file)
        nbfal = mafft_align(num_fam, prt_file, mafft_file, nbfprt, logger)
    # If problem with alignment, return False
    if not nbfal:
        return False
    # If btr file already exists, means that it was already done before, and not removed because
    # extractions and mafft files are ok. So, return True, saying that btr file is done,
    # next step will be to check it, add missing genomes etc.
    if os.path.isfile(btr_file):
        message = ("fam {}: Will redo back-translation, because found a different number of "
                   "proteins aligned in {} ({}) and genes back-translated in "
                   "existing {}").format(num_fam, mafft_file, nbfal, btr_file)
        res = check_nb_seqs(btr_file, [nbfal, ngenomes], logger, message)
        if not res:
            utils.remove(btr_file)
        else:
            return "OK"
    # If btr file does not exist (removed because problem with mafft generated before,
    # or just not generated yet), do back-translation, and return True if it went well,
    # False otherwise
    return back_translate(num_fam, mafft_file, gen_file, btr_file, nbfal, logger)


def check_extractions(num_fam, miss_file, prt_file, gen_file, ngenomes, logger):
    """
    Check that extractions went well for the given family:

    - check number of proteins and genes extracted compared to the
      number of genomes

    Parameters
    ----------
    num_fam : int
        current family number
    miss_file : str
        path to file containing the list of genomes missing for the current family
    prt_file : str
        path to file containing all proteins extracted
    gen_file : str
        path to file containing all genes extracted
    ngenomes : int
        total number of genomes in dataset
    logger : logging.Logger
        logger with queueHandler to give logs to main logger

    Returns
    -------
    bool or int
        False if any problem (nbmiss+prt != nbgenomes or nbmiss+gen != nbgenomes). If no
        problem, returns the number of proteins/genes extracted
    """
    logger.log(utils.detail_lvl(), "Checking extractions for family {}".format(num_fam))

    # Check that extractions went well
    nbmiss = utils.count(miss_file)
    nbfprt = utils.grep(prt_file, "^>", counts=True)
    nbfgen = utils.grep(gen_file, "^>", counts=True)
    if nbmiss + nbfprt != ngenomes:
        logger.error(("fam {}: wrong sum of missing genomes ({}) and prt extracted ({}) for {} "
                      "genomes in the dataset.").format(num_fam, nbmiss, nbfprt, ngenomes))
        return False
    if nbmiss + nbfgen != ngenomes:
        logger.error(("fam {}: wrong sum of missing genomes ({}) and gen extracted ({}) for {} "
                      "genomes in the dataset.").format(num_fam, nbmiss, nbfgen, ngenomes))
        return False
    return nbfprt


def mafft_align(num_fam, prt_file, mafft_file, nbfprt, logger):
    """
    Align all proteins of the given family with mafft

    Parameters
    ----------
    num_fam : int
        current family number
    prt_file : str
        path to file containing all proteins extracted
    mafft_file : str
        path to file which will contain proteins alignment
    nbfprt : int
        number of proteins extracted in prt file
    logger : logging.Logger
        logger with queueHandler to give logs to main logger

    Returns
    -------
    bool
        True if no problem (alignment ok, same number of proteins extracted and aligned),
        False otherwise
    """
    logger.log(utils.detail_lvl(), "Aligning family {}".format(num_fam))
    cmd = "mafft --auto {}".format(prt_file)
    error = "Problem while trying to align fam {}".format(num_fam)
    stdout = open(mafft_file, "w")
    stderr = open(mafft_file + ".log", "w")
    logger.log(utils.detail_lvl(), cmd)
    ret = utils.run_cmd(cmd, error, stdout=stdout, stderr=stderr, logger=logger)
    stdout.close()
    if not isinstance(ret, int):
        ret = ret.returncode
    if ret != 0:
        os.remove(mafft_file)
        return False
    message = ("fam {}: different number of proteins extracted in {} ({}) and proteins "
               "aligned in {}").format(num_fam, prt_file, nbfprt, mafft_file)
    return check_nb_seqs(mafft_file, nbfprt, logger, message)


def back_translate(num_fam, mafft_file, gen_file, btr_file, nbfal, logger):
    """
    Backtranslate protein alignment to nucleotides

    Parameters
    ----------
    num_fam : int
        current family number. Used for log messages
    mafft_file : str
        path to file containing protein alignments by mafft
    gen_file : str
        path to file containing all sequences, not aligned, in nucleotides. It is used to
        convert the alignment in proteins into a nucleotide alignment
    btr_file : str
        path to the file that will contain the nucleotide alignment
    nbfal : int
        number of sequences aligned for the family by mafft
    logger : logging.Logger
        logger with queueHandler to give logs to main logger

    Returns
    -------
    bool
        - False if problem (back-translation, different number of families...)
        - number of sequences in btr file if everything went well
    """
    logger.log(utils.detail_lvl(), "Back-translating family {}".format(num_fam))
    curpath = os.path.dirname(os.path.abspath(__file__))
    awk_script = os.path.join(curpath, "prt2codon.awk")
    cmd = "awk -f {} {} {}".format(awk_script, mafft_file, gen_file)
    stdout = open(btr_file, "w")
    error = "Problem while trying to backtranslate {} to a nucleotide alignment".format(mafft_file)
    ret = utils.run_cmd(cmd, error, stdout=stdout, logger=logger)
    stdout.close()
    if not isinstance(ret, int):
        ret = ret.returncode
    if ret != 0:
        os.remove(btr_file)
        return False
    message = ("fam {}: different number of proteins aligned in {} ({}) and genes "
               "back-translated in {}").format(num_fam, mafft_file, nbfal, btr_file)
    return check_nb_seqs(mafft_file, nbfal, logger, message)


def check_nb_seqs(alnfile, nbfal, logger, message):
    """
    Check the number of sequences in the given alignment file

    Parameters
    ----------
    alnfile : str
        path to alignment file
    nbfal : int or []
        expected number of sequences, or list of expected numbers of sequences
    logger : logging.Logger
        logger with queueHandler to give logs to main logger
    message : str
        message to write when sequence numbers do not match

    Returns
    -------
    bool or int
        - False if not same number of sequences
        - nbseqs in align file if found among values in 'nbfal'
    """
    nbseqs = utils.grep(alnfile, "^>", counts=True)
    if isinstance(nbfal, int):
        nbfal = [nbfal]
    for num in nbfal:
        if nbseqs == num:
            return nbseqs

    logger.error(message + ' ({})'.format(nbseqs))
    return False


def check_lens(aln_file, num_fam, logger):
    """
    In the given alignment file, check that all sequences have the same length.
    If there is no problem, it returns the length of alignment, and the number
    of sequences aligned.

    Parameters
    ----------
    aln_file : str
        path to the alignment file to check
    num_fam : int
        current family number. Used for log message if problem
    logger : logging.Logger
        logger having a queue Handler to give logs to the main logger in the main process

    Returns
    -------
    bool or tuple
        False if there is a problem in the alignment (sequences do not all have the
        same length).
        If they all have the same length, returns this length and the number of sequences
    """
    nb_gen = 0
    all_sums = set()
    with open(aln_file, "r") as btrf:
        cur_sum = 0
        start = True
        for line in btrf:
            if line.startswith(">"):
                nb_gen += 1
                if not start:
                    all_sums.add(cur_sum)
                    cur_sum = 0
            else:
                start = False
                cur_sum += len(line.strip())
        all_sums.add(cur_sum)
    if len(all_sums) > 1:
        logger.error(("alignments for family {} do not all have the same "
                      "length. Lengths found are: {}\n").format(num_fam, all_sums))
        return False
    return list(all_sums)[0], nb_gen



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the post_align submodule in align module
"""
import os
import pytest
import shutil
import logging

import PanACoTA.align_module.post_align as pal
import test.test_unit.utilities_for_tests as tutil
from PanACoTA import utils

# Define common variables
ALPATH = os.path.join("test", "data", "align")
EXPPATH = os.path.join(ALPATH, "exp_files")
TESTPATH = os.path.join(ALPATH, "test_files")
GENEPATH = os.path.join(ALPATH, "generated_by_unit-tests")
LOGFILE_BASE = "logs_test_postalign"
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]

@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    utils.init_logger(LOGFILE_BASE, 0, 'test_postalign', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    print("teardown")


def test_get_genome():
    """
    Given a header and a list of genomes, check that it returns the expected genome
    """
    header = ">TOTO.0215.00002.i006_00065"
    genomes = ["TOTO.0315.00001", "ESCO.0215.00002", "ESCO.0215.00001", "TOTO.0215.00002"]
    assert pal.get_genome(header, genomes) == "TOTO.0215.00002"


def test_get_genome_not_start():
    """
    Given a header and a list of genomes, check that it returns the expected genome. The genome
    name is not at the beginning of the protein name
    """
    header = ">mongenome,TOTO.0215.00002.i006_00065"
    genomes = ["TOTO.0315.00001", "ESCO.0215.00002", "ESCO.0215.00001", "TOTO.0215.00002"]
    assert pal.get_genome(header, genomes) == "TOTO.0215.00002"


def test_get_genome_notfound(caplog):
    """
    Test that when header given, but no genome of the list corresponds to it, it exits with an
    error message.
    """
    header = ">TOTO.0215.00002.i006_00065"
    genomes = ["TOTO.0315.00001", "ESCO.0215.00002", "ESCO.0215.00001"]
    assert pal.get_genome(header, genomes) is None
    assert ("Protein TOTO.0215.00002.i006_00065 does not correspond to any genome name given... "
            "['TOTO.0315.00001', 'ESCO.0215.00002', 'ESCO.0215.00001']") in caplog.text


def test_write_groups(caplog):
    """
    Check that giving a list of genomes and corresponding sequences, it writes all of them in
    fasta format in the given output file.
    """
    caplog.set_level(logging.DEBUG)
    outfile = os.path.join(GENEPATH, "test_write_groups.fa")
    sequences = {"ESCO.0216.00001": ["AAAAA-", "TTTTT", "CCCCC", "GGGGG"],
                 "ESCO.0216.00002": ["AAAAT-", "TTTTA", "-CCCG", "GGGGC"],
                 "ESCO.0216.00003": ["AAAAAA", "TT-TT", "CCC-C", "GCGG-"]}
    pal.write_groups(outfile, sequences)
    assert "Writing alignments per genome" in caplog.text
    with open(outfile, "r") as outf:
        lines = outf.readlines()
    assert len(lines) == 6
    exp_lines = [">ESCO.0216.00001\n", "AAAAA-TTTTTCCCCCGGGGG\n",
                 ">ESCO.0216.00002\n", "AAAAT-TTTTA-CCCGGGGGC\n",
                 ">ESCO.0216.00003\n", "AAAAAATT-TTCCC-CGCGG-\n"]
    assert lines == exp_lines


def test_read_alignment(caplog):
    """
    Giving a file with proteins aligned, and a list of genomes, returns, for each genome,
    the list of sequences found.
    """
    caplog.set_level(logging.DEBUG)
    alnfile = os.path.join(TESTPATH, "complete.cat.fictive4genomes.aln")
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    seqs = pal.read_alignments(alnfile, all_genomes)
    exp_seq = {"GEN2.1017.00001": ["AAATCTCGGAGATCTCGCGCGGATATATCGCGCTATAGCGCT",
                                   "---TTCGCGCGGAGAGGCGCTCTCGGCG",
                                   "----------------------------------------------------"],
               "GEN4.1111.00001": ["AAATCTCGGAGTTCTCGCGCGGGTA---CGCGCTATAGCGCT",
                                   "---TTCGCGCGGAGAGGCGCTCTCG---",
                                   "CGGTATAGGGCGCTAC----CCGGATATAGGGCTCTCGGAATTCGCAAC-CC"],
               "GENO.1017.00001": ["-------GGAGATCTCGCGCGGATATATCGCGCTATAGCGCT",
                                   "AAATTCGCGCGGAGAGGCGCTCTCGAAA",
                                   "CGGTATAGGGCGCTACTTTTCCGGATATAGGGCTCTCGGAATTCG--ACACC"],
               "GENO.1216.00002": ["AAATCTCGGAGATCTCGCGCGGATATATAAAACTATAGCGCT",
                                   "----------------------------",
                                   "CGGTATGGGGCGCTAC----CCGGATATAGGGCTCTCGGAATTCGCAAC-CG"]
               }
    assert seqs == exp_seq
    assert "3 sequences found per genome" in caplog.text


def test_read_alignment_diffnbseq(caplog):
    """
    Giving a file with proteins aligned, and a list of genomes, with 3 proteins for each genome,
    except for 1 genome which has 4 proteins. Exits with an error message.
    """
    caplog.set_level(logging.DEBUG)
    alnfile = os.path.join(TESTPATH, "complete.cat.fictive4genomes-diffnbseq.aln")
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    assert pal.read_alignments(alnfile, all_genomes) is None
    assert ("Problems occurred while grouping alignments by genome: all genomes do not have the "
            "same number of sequences. Check that each protein name contains the name of the "
            "genome from which it comes.") in caplog.text


def test_read_alignment_nogenome(caplog):
    """
    Giving a file with proteins aligned, and a list of genomes, but with one genome missing,
    check that it returns None, and an error message specifying the protein not having a
    corresponding genome.
    """
    caplog.set_level(logging.DEBUG)
    alnfile = os.path.join(TESTPATH, "complete.cat.fictive4genomes-diffnbseq.aln")
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001"]
    assert pal.read_alignments(alnfile, all_genomes) is None
    assert ("Protein GENO.1216.00002.i0001_00003 does not correspond to any genome name given... "
            "['GEN2.1017.00001', 'GEN4.1111.00001', 'GENO.1017.00001']") in caplog.text


def test_group_by_genome(caplog):
    """
    Test that giving a file with all proteins aligned, a list of genomes, and an output
    filename, it writes in output the alignment grouped by genome and returns True
    """
    caplog.set_level(logging.DEBUG)
    alnfile = os.path.join(TESTPATH, "complete.cat.fictive4genomes.aln")
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    outgrp = os.path.join(GENEPATH, "test_group_by_genome")
    args = (all_genomes, alnfile, outgrp)
    assert pal.group_by_genome(args)
    exp_grp = os.path.join(EXPPATH, "exp_fictive.grp.aln")
    assert tutil.compare_order_content(outgrp, exp_grp)
    assert "3 sequences found per genome" in caplog.text
    assert "Writing alignments per genome" in caplog.text


def test_group_by_genome_nogenome(caplog):
    """
    Test that giving a file with all proteins aligned, a list of genomes but with 1 genome missing,
    it returns False, and an error message with the protein not having a corresponding genome
    """
    caplog.set_level(logging.DEBUG)
    alnfile = os.path.join(TESTPATH, "complete.cat.fictive4genomes.aln")
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001"]
    outgrp = os.path.join(GENEPATH, "test_group_by_genome")
    args = (all_genomes, alnfile, outgrp)
    assert pal.group_by_genome(args) is False
    assert not os.path.isfile(outgrp)
    assert ("Protein GENO.1216.00002.i0001_00003 does not correspond to any genome name given... "
            "['GEN2.1017.00001', 'GEN4.1111.00001', 'GENO.1017.00001']") in caplog.text


def test_launch_gbg(caplog):
    """
    Giving an alignment file, and a list of genomes, and a tree directory, check that it
    generates the expected filename, with the expected content, and returns True.
    """
    caplog.set_level(logging.DEBUG)
    all_genomes = ["GENO.1216.00002", "GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001"]
    alnfile = os.path.join(TESTPATH, "complete.cat.pers4genomes.aln")
    status = "Done"
    treedir = os.path.join(GENEPATH, "test_phylo-pers4genomes")
    # Create tree dir, with empty output file
    os.makedirs(treedir)
    out_grp = os.path.join(treedir, "TESTlaunch.grp.aln")
    open(out_grp, "w").close()
    dname = "TESTlaunch"
    quiet = False
    assert pal.launch_group_by_genome(all_genomes, alnfile, status, out_grp, dname, quiet) is True
    exp_grp = os.path.join(EXPPATH, "exp_pers4genomes.grp.aln")
    assert tutil.compare_order_content(out_grp, exp_grp)
    assert "Grouping alignments per genome" in caplog.text


def test_launch_gbg_existsempty(caplog):
    """
    Giving an alignment file, and a list of genomes, and saying that we did not re-create this
    concatenated file. As the group by genome file already exists (even if empty...), it won't
    recreate it, just return None and keep it.
    """
    caplog.set_level(logging.DEBUG)
    all_genomes = ["GENO.1216.00002", "GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001"]
    alnfile = os.path.join(TESTPATH, "complete.cat.pers4genomes.aln")
    status = "OK"
    treedir = os.path.join(GENEPATH, "test_phylo-pers4genomes")
    # Create tree dir, with empty output file
    os.makedirs(treedir)
    out_grp = os.path.join(treedir, "TESTlaunch.grp.aln")
    open(out_grp, "w").close()
    dname = "TESTlaunch"
    quiet = False
    assert pal.launch_group_by_genome(all_genomes, alnfile, status, out_grp, dname, quiet) is True
    assert "Grouping alignments per genome" not in caplog.text
    with open(out_grp, "r") as outf:
        assert outf.readlines() == []


def test_launch_gbg_ok_notexist(caplog):
    """
    Giving an alignment file, and a list of genomes, and a tree directory, and saying that we
    did not re-create this concatenated file. As group by genome file does not exist,
    it will create it as if concatenation was just done, and return True
    """
    caplog.set_level(logging.DEBUG)
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    alnfile = os.path.join(TESTPATH, "complete.cat.pers4genomes.aln")
    status = "OK"
    treedir = os.path.join(GENEPATH, "test_phylo-pers4genomes")
    # Create tree dir
    os.makedirs(treedir)
    out_grp = os.path.join(treedir, "TESTlaunch.grp.aln")
    dname = "TESTlaunch"
    quiet = False
    assert pal.launch_group_by_genome(all_genomes, alnfile, status, out_grp, dname, quiet) is True
    exp_grp = os.path.join(EXPPATH, "exp_pers4genomes.grp.aln")
    assert tutil.compare_order_content(out_grp, exp_grp)
    assert "Grouping alignments per genome" in caplog.text


def test_launch_gbg_nogenome(caplog):
    """
    Giving an alignment file, and a list of genomes with one genome missing, check that it
    returns False, with an error message for the protein not having a corresponding genome,
    and does not create grp file
    """
    caplog.set_level(logging.DEBUG)
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001"]
    alnfile = os.path.join(TESTPATH, "complete.cat.pers4genomes.aln")
    status = "OK"
    treedir = os.path.join(GENEPATH, "test_phylo-pers4genomes")
    # Create tree dir
    os.makedirs(treedir)
    out_grp = os.path.join(treedir, "TESTlaunch.grp.aln")
    dname = "TESTlaunch"
    quiet = False
    assert pal.launch_group_by_genome(all_genomes, alnfile, status, treedir, dname, quiet) is False
    assert not os.path.isfile(out_grp)
    assert "Grouping alignments per genome" in caplog.text


def test_concat(caplog):
    """
    Given a list of families, and a directory where are alignment files, check that the files
    corresponding to the given families are concatenated as expected, and it returns "Done" and
    expected output filename
    """
    caplog.set_level(logging.DEBUG)
    # Prepare aldir with all needed alignment files
    aldir = os.path.join(GENEPATH, "test_concat_aldir")
    dname = "TEST-concat"
    prefix = os.path.join(aldir, dname)
    orig_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    orig_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    orig_btr11 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.11.aln")
    btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    btr11 = os.path.join(aldir, dname + "-mafft-prt2nuc.11.aln")
    os.makedirs(aldir)
    shutil.copyfile(orig_btr1, btr1)
    shutil.copyfile(orig_btr8, btr8)
    shutil.copyfile(orig_btr11, btr11)
    # Other parameters, and run concatenation
    fam_nums = [1, 8, 11]
    quiet = False
    output, mess = pal.concat_alignments(fam_nums, prefix, quiet)
    assert output == os.path.join(aldir, dname + "-complete.cat.aln")
    ref_concat = os.path.join(EXPPATH, "exp_concat_4genomes-fam1-8-11.aln")
    assert tutil.compare_order_content(output, ref_concat)
    assert mess == "Done"
    assert "Concatenating all alignment files" in caplog.text


def test_concat_quiet(caplog):
    """
    Given a list of families, and a directory where are alignment files, check that the files
    corresponding to the given families are concatenated as expected, and it returns "Done" and
    expected output filename
    """
    caplog.set_level(logging.DEBUG)
    # Prepare aldir with all needed alignment files
    aldir = os.path.join(GENEPATH, "test_concat_aldir")
    dname = "TESTconcat"
    prefix = os.path.join(aldir, dname)
    orig_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    orig_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    orig_btr11 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.11.aln")
    btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    btr11 = os.path.join(aldir, dname + "-mafft-prt2nuc.11.aln")
    os.makedirs(aldir)
    shutil.copyfile(orig_btr1, btr1)
    shutil.copyfile(orig_btr8, btr8)
    shutil.copyfile(orig_btr11, btr11)
    # Other parameters, and run concatenation
    fam_nums = [1, 8, 11]
    quiet = True
    output, mess = pal.concat_alignments(fam_nums, prefix, quiet)
    assert output == os.path.join(aldir, dname + "-complete.cat.aln")
    ref_concat = os.path.join(EXPPATH, "exp_concat_4genomes-fam1-8-11.aln")
    assert tutil.compare_order_content(output, ref_concat)
    assert mess == "Done"
    assert "Concatenating all alignment files" in caplog.text


def test_concat_noalignfile(caplog):
    """
    Given a list of families, and a directory where are alignment files except for 1 family,
    for which we do not have the alignment file
    """
    caplog.set_level(logging.DEBUG)
    # Prepare aldir with all needed alignment files
    aldir = os.path.join(GENEPATH, "test_concat_aldir")
    dname = "TESTconcat"
    prefix = os.path.join(aldir, dname)
    orig_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    orig_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    orig_btr11 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.11.aln")
    btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    btr11 = os.path.join(aldir, dname + "-mafft-prt2nuc.11.aln")
    os.makedirs(aldir)
    shutil.copyfile(orig_btr1, btr1)
    shutil.copyfile(orig_btr8, btr8)
    shutil.copyfile(orig_btr11, btr11)
    # Other parameters, and run concatenation
    fam_nums = [1, 8, 11, 15]
    quiet = False
    with pytest.raises(SystemExit):
        pal.concat_alignments(fam_nums, prefix, quiet)
    assert ("The alignment file "
            "test/data/align/generated_by_unit-tests/test_concat_aldir/"
            "TESTconcat-mafft-prt2nuc.15.aln "
            "does not exist. Please check the families you want, and their corresponding "
            "alignment files") in caplog.text


def test_concat_outexists(caplog):
    """
    Given a list of families, and a directory where are alignment files and the concatenated
    file, check that it keeps this outfile, and returns "OK" with a warning message
    """
    caplog.set_level(logging.DEBUG)
    # Prepare aldir with all needed alignment files
    aldir = os.path.join(GENEPATH, "test_concat_aldir")
    dname = "TESTconcat"
    prefix = os.path.join(aldir, dname)
    orig_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    orig_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    orig_btr11 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.11.aln")
    btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    btr11 = os.path.join(aldir, dname + "-mafft-prt2nuc.11.aln")
    os.makedirs(aldir)
    shutil.copyfile(orig_btr1, btr1)
    shutil.copyfile(orig_btr8, btr8)
    shutil.copyfile(orig_btr11, btr11)
    # Create empty concatenated file
    outempty = os.path.join(aldir, dname + "-complete.cat.aln")
    open(outempty, "w").close()
    # Other parameters, and run concatenation
    fam_nums = [1, 8, 11]
    quiet = False
    output, mess = pal.concat_alignments(fam_nums, prefix, quiet)
    assert output == outempty
    assert mess == "OK"
    with open(output, "r") as outf:
        assert outf.readlines() == []
    assert "Alignments already concatenated" in caplog.text
    assert ("Alignments already concatenated in "
            "test/data/align/generated_by_unit-tests/test_concat_aldir/"
            "TESTconcat-complete.cat.aln. "
            "Program will use it for next steps") in caplog.text


def test_postalign(caplog):
    """
    Test that when running post-alignment on a folder containing all expected alignment files,
    it creates concatenated alignments, and a folder Phylo with the alignments grouped by genome.
    """
    caplog.set_level(logging.DEBUG)
    # define parameters
    fam_nums = [1, 8, 11]
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    outdir = os.path.join(GENEPATH, "test_post-align")
    aldir = os.path.join(outdir, "aldir_post-align")
    os.makedirs(aldir)
    dname = "TESTpost"
    prefix = os.path.join(aldir, dname)
    quiet = False
    # Prepare aldir with all needed alignment files
    orig_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    orig_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    orig_btr11 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.11.aln")
    btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    btr11 = os.path.join(aldir, dname + "-mafft-prt2nuc.11.aln")
    shutil.copyfile(orig_btr1, btr1)
    shutil.copyfile(orig_btr8, btr8)
    shutil.copyfile(orig_btr11, btr11)
    # Run post-alignment
    pal.post_alignment(fam_nums, all_genomes, prefix, outdir, dname, quiet)
    # Check that concatenated file is created and with expected content
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    assert os.path.isfile(out_concat)
    ref_concat = os.path.join(EXPPATH, "exp_concat_4genomes-fam1-8-11.aln")
    assert tutil.compare_order_content(out_concat, ref_concat)
    # Check that grouped by genome file is created, with expected content
    treedir = os.path.join(outdir, "Phylo-" + dname)
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    assert os.path.isfile(out_grp)
    exp_grp = os.path.join(EXPPATH, "exp_grp_4genomes-fam1-8-11.aln")
    assert tutil.compare_order_content(out_grp, exp_grp)
    # check logs
    assert "Concatenating all alignment files" in caplog.text
    assert "Grouping alignments per genome" in caplog.text


def test_postalign_missalign(caplog):
    """
    Test that when running post-alignment on a folder containing all expected alignment files
    except 1, it exits with an error message indicating the missing alignment file.
    """
    caplog.set_level(logging.DEBUG)
    # define parameters
    fam_nums = [1, 8, 11]
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    outdir = os.path.join(GENEPATH, "test_post-align_missalign")
    aldir = os.path.join(outdir, "aldir_post-align")
    os.makedirs(aldir)
    dname = "TESTpost"
    prefix = os.path.join(aldir, dname)
    quiet = False
    # Prepare aldir with all needed alignment files
    orig_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    orig_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    shutil.copyfile(orig_btr1, btr1)
    shutil.copyfile(orig_btr8, btr8)
    # Run post-alignment
    with pytest.raises(SystemExit):
        pal.post_alignment(fam_nums, all_genomes, prefix, outdir, dname, quiet)
    assert ("The alignment file test/data/align/generated_by_unit-tests/"
            "test_post-align_missalign/aldir_post-align/TESTpost-mafft-prt2nuc.11.aln "
            "does not exist. Please check the families you want, and their corresponding "
            "alignment files") in caplog.text
    # Check that concatenated file is not created
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    assert not os.path.isfile(out_concat)


def test_postalign_missgenome(caplog):
    """
    Test that when running post-alignment on a folder containing all expected alignment files,
    but giving incomplete list of genomes, it exits with error message specifying protein which
    does not belong to any given genome name.
    """
    caplog.set_level(logging.DEBUG)
    # define parameters
    fam_nums = [1, 8, 11]
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001"]
    outdir = os.path.join(GENEPATH, "test_post-align")
    aldir = os.path.join(outdir, "aldir_post-align")
    os.makedirs(aldir)
    dname = "TESTpost"
    prefix = os.path.join(aldir, dname)
    quiet = False
    # Prepare aldir with all needed alignment files
    orig_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    orig_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    orig_btr11 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.11.aln")
    btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    btr11 = os.path.join(aldir, dname + "-mafft-prt2nuc.11.aln")
    shutil.copyfile(orig_btr1, btr1)
    shutil.copyfile(orig_btr8, btr8)
    shutil.copyfile(orig_btr11, btr11)
    # Run post-alignment
    pal.post_alignment(fam_nums, all_genomes, prefix, outdir, dname, quiet)
    # Check that concatenated file is created and with expected content
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    assert os.path.isfile(out_concat)
    ref_concat = os.path.join(EXPPATH, "exp_concat_4genomes-fam1-8-11.aln")
    assert tutil.compare_order_content(out_concat, ref_concat)
    # Check that grouped by genome file is not created
    treedir = os.path.join(outdir, "Phylo-" + dname)
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    assert not os.path.isfile(out_grp)
    # check logs
    assert "Concatenating all alignment files" in caplog.text
    assert "Grouping alignments per genome" in caplog.text
    assert "An error occurred. We could not group sequences by genome." in caplog.text


def test_postalign_diffnbseq(caplog):
    """
    Test that when running post-alignment on a folder containing all expected alignment files,
    but giving incomplete list of genomes, it exits with error message specifying protein which
    does not belong to any given genome name.
    """
    caplog.set_level(logging.DEBUG)
    # define parameters
    fam_nums = [1, 8, 11]
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001"]
    outdir = os.path.join(GENEPATH, "test_post-align")
    aldir = os.path.join(outdir, "aldir_post-align")
    os.makedirs(aldir)
    dname = "TESTpost"
    prefix = os.path.join(aldir, dname)
    quiet = False
    # Prepare aldir with all needed alignment files
    orig_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    orig_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    orig_btr11 = os.path.join(TESTPATH, "error-mafft-prt2nuc.11.sup-seq.aln")
    btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    btr11 = os.path.join(aldir, dname + "-mafft-prt2nuc.11.aln")
    shutil.copyfile(orig_btr1, btr1)
    shutil.copyfile(orig_btr8, btr8)
    shutil.copyfile(orig_btr11, btr11)
    # Run post-alignment
    pal.post_alignment(fam_nums, all_genomes, prefix, outdir, dname, quiet)
    # Check that concatenated file is created and with expected content
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    assert os.path.isfile(out_concat)
    ref_concat = os.path.join(TESTPATH, "error-concat-sup-seq.aln")
    assert tutil.compare_order_content(out_concat, ref_concat)
    # Check that grouped by genome file is not created
    treedir = os.path.join(outdir, "Phylo-" + dname)
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    assert not os.path.isfile(out_grp)
    # check logs
    assert "Concatenating all alignment files" in caplog.text
    assert "Grouping alignments per genome" in caplog.text
    assert "An error occurred. We could not group sequences by genome." in caplog.text



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for general_format_functions.py
"""

import os
import logging
import shutil
from io import StringIO
import pytest

import test.test_unit.utilities_for_tests as tutil
import PanACoTA.utils as utils
from PanACoTA.annotate_module import format_prodigal as prodigalfunc

ANNOTEDIR = os.path.join("test", "data", "annotate")
GENOMES_DIR = os.path.join(ANNOTEDIR, "genomes")
EXP_ANNOTE = os.path.join(ANNOTEDIR, "exp_files")
TEST_ANNOTE = os.path.join(ANNOTEDIR, "test_files")
GENEPATH = os.path.join(ANNOTEDIR, "generated_by_unit-tests")

LOGFILE_BASE = os.path.join(GENEPATH, "logs", "logfile")
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    os.mkdir(GENEPATH)
    # utils.init_logger(LOGFILE_BASE, 0, 'test_fastme', verbose=1)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    # for f in LOGFILES:
    #     if os.path.exists(f):
    #         os.remove(f)
    print("teardown")


def test_create_gen_lst(caplog):
    """
    Check that generated gen and lst files are as expected.
    In the test file, all genomes have names different from gembase name
    This test file contains the following aspects:
    - gene in D strand (start < end)
    - gene in C strand (start > end)
    - CDS features
    - contigs with more than 2 genes
    - contig with only 2 genes (both 'b' loc)
    - contig with 1 gene ('b' loc)
    - contig without gene (should be skipped)
    """
    caplog.set_level(logging.DEBUG)
    genfile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                           "prodigal.outtest.ok.ffn")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007"
               }
    name = "test.0417.00002"
    res_gen_file = os.path.join(GENEPATH, "prodigal_res.gen")
    res_lst_file = os.path.join(GENEPATH, "prodigal_res.lst")
    gpath = "original_genome_name"
    assert prodigalfunc.create_gene_lst(contigs, genfile, res_gen_file, res_lst_file, gpath,
                                        name)
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    assert tutil.compare_order_content(exp_lst, res_lst_file)
    exp_gen = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.gen")
    assert tutil.compare_order_content(exp_gen, res_gen_file)


def test_create_gen_lst_cont_unknown(caplog):
    """
    A contig name in the gen file does not exist -> error message, and all result files
    must be removed for this genome
    """
    caplog.set_level(logging.DEBUG)
    genfile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                           "prodigal.outtest.ok.faa")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007"
               }
    name = "test.0417.00002"
    res_gen_file = os.path.join(GENEPATH, "prodigal_res.gen")
    res_lst_file = os.path.join(GENEPATH, "prodigal_res.lst")
    gpath = "original_genome_name"
    assert not prodigalfunc.create_gene_lst(contigs, genfile, res_gen_file, res_lst_file,
                                            gpath, name)
    assert ("'my_contig' found in test/data/annotate/test_files/original_name.fna-prodigalRes/"
            "prodigal.outtest.ok.faa does not exist in original_genome_name") in caplog.text


def test_create_gff(caplog):
    """
    Check generated gff file. Must have all sequences in header (even replicons without gene),
    and then 1 line per gene
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    gfffile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                           "prodigal.outtest.ok.gff")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007"
               }
    sizes = {"test.0417.00002.0001": 84,
             "test.0417.00002.0002": 103,
             "test.0417.00002.0003": 122,
             "test.0417.00002.0004": 35,
             "test.0417.00002.0005": 198,
             "test.0417.00002.0006": 128,
             "test.0417.00002.0007": 85,
            }
    res_gff_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    gpath = "original_genome_name"
    assert prodigalfunc.create_gff(gpath, gfffile, res_gff_file, exp_lst, contigs, sizes)
    exp_gff = os.path.join(EXP_ANNOTE, "res_create_gff_prodigal.gff")
    assert tutil.compare_order_content(exp_gff, res_gff_file)


def test_create_gff_wrong_start(caplog):
    """
    Check that when trying to generate gff, if problem, exits with expected message
    Here, the start position in the gff file generated by prodigal is not the same as
    the start position in the lstinfo file (which was taken from the ffn
    file generated by prodigal)
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    gfffile = os.path.join(TEST_ANNOTE, "prodigal.outtest.wrong-start.gff")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007"
               }
    sizes = {"test.0417.00002.0001": 14000,
             "test.0417.00002.0002": 5000,
             "test.0417.00002.0003": 4600,
             "test.0417.00002.0004": 8000,
             "test.0417.00002.0005": 1,
             "test.0417.00002.0006": 10,
             "test.0417.00002.0007": 15000,
            }
    name = "test.0417.00002"
    res_gff_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    gpath = "original_genome_name"
    assert not prodigalfunc.create_gff(gpath, gfffile, res_gff_file, exp_lst, contigs, sizes)
    assert ("Files prodigal.outtest.wrong-start.ffn and "
            "prodigal.outtest.wrong-start.gff "
            "(in prodigal tmp_files: original_genome_name-prodigalRes) "
            "do not have the same start value for gene EPKOMDHM_00008 "
            "(78 in gff, 77 in ffn") in caplog.text


def test_create_gff_wrong_end(caplog):
    """
    Check that when trying to generate gff, if problem, exits with expected message
    Here, the start position in the gff file generated by prodigal is not the same as
    the start position in the lstinfo file (which was taken from the ffn
    file generated by prodigal)
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    gfffile = os.path.join(TEST_ANNOTE, "prodigal.outtest.wrong-end.gff")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007"
               }
    sizes = {"test.0417.00002.0001": 14000,
             "test.0417.00002.0002": 5000,
             "test.0417.00002.0003": 4600,
             "test.0417.00002.0004": 8000,
             "test.0417.00002.0005": 1,
             "test.0417.00002.0006": 10,
             "test.0417.00002.0007": 15000,
            }
    res_gff_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    gpath = "original_genome_name"
    assert not prodigalfunc.create_gff(gpath, gfffile, res_gff_file, exp_lst, contigs, sizes)
    assert ("Files prodigal.outtest.wrong-end.ffn and "
            "prodigal.outtest.wrong-end.gff "
            "(in prodigal tmp_files: original_genome_name-prodigalRes) "
            "do not have the same end value for gene EPKOMDHM_00009 "
            "(2347 in gff, 2346 in ffn") in caplog.text


def test_create_gff_wrong_type(caplog):
    """
    Check that when trying to generate gff, if problem, exits with expected message
    Here, the start position in the gff file generated by prodigal is not the same as
    the start position in the lstinfo file (which was taken from the ffn
    file generated by prodigal)
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    gfffile = os.path.join(TEST_ANNOTE, "prodigal.outtest.wrong-type.gff")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007"
               }
    sizes = {"test.0417.00002.0001": 14000,
             "test.0417.00002.0002": 5000,
             "test.0417.00002.0003": 4600,
             "test.0417.00002.0004": 8000,
             "test.0417.00002.0005": 1,
             "test.0417.00002.0006": 10,
             "test.0417.00002.0007": 15000,
            }
    res_gff_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    gpath = "original_genome_name"
    assert not prodigalfunc.create_gff(gpath, gfffile, res_gff_file, exp_lst, contigs, sizes)
    assert ("Files prodigal.outtest.wrong-type.ffn and "
            "prodigal.outtest.wrong-type.gff "
            "(in prodigal tmp_files: original_genome_name-prodigalRes) "
            "do not have the same type value for gene EPKOMDHM_00008 "
            "(tRNA in gff, CDS in ffn") in caplog.text


def test_create_prt(caplog):
    """
    Check that prt file is generated as expected
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                           "prodigal.outtest.ok.faa")
    res_prt_file = os.path.join(GENEPATH, "prodigal_res.prt")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    assert prodigalfunc.create_prt(protfile, res_prt_file, exp_lst)
    exp_prt = os.path.join(EXP_ANNOTE, "res_create_prt_prodigal.faa")
    assert tutil.compare_order_content(exp_prt, res_prt_file)


def test_create_prt_wrong_lst(caplog):
    """
    Check that prt file is not generated if there is a problem in the lst file
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                           "prodigal.outtest.ok.faa")
    res_prt_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(TEST_ANNOTE, "test_create_prt_prodigal-wrongformat.lst")
    assert not prodigalfunc.create_prt(protfile, res_prt_file, exp_lst)
    assert ("Problem in format of lstline (1279\t2346\tCDS\ttest.0417.00002.0002i_00005\tNA\t"
            "| NA | NA | fefer | NA") in caplog.text


def test_create_prt_short_lst(caplog):
    """
    Check that prt file is not generated when there are more proteins in faa than in lst
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                           "prodigal.outtest.ok.faa")
    res_prt_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(TEST_ANNOTE, "test_create_prt_prodigal-shortlst.lst")
    assert not prodigalfunc.create_prt(protfile, res_prt_file, exp_lst)
    assert ("No more protein in lst file. We cannot get information on this "
            "protein (>toto_00011 # 2419 # 3000 # 1 # a)! Check that you do not have "
            "more proteins than genes in prodigal results") in caplog.text


def test_create_prt_end_not_int_lst(caplog):
    """
    Check that prt file is not generated when an end value of a protein in lst file is not an int
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                           "prodigal.outtest.ok.faa")
    res_prt_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(TEST_ANNOTE, "test_create_prt_prodigal-notint.lst")
    assert not prodigalfunc.create_prt(protfile, res_prt_file, exp_lst)
    assert ("Start and/or end of protein test.0417.00002.0001i_00002 position is not a number "
            "(start = 4416; end = a6068)") in caplog.text


def test_create_prt_not_divisible3_lst(caplog):
    """
    Check that prt file is not generated when there is a gene which length is not
    divisible by 3
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                           "prodigal.outtest.ok.faa")
    res_prt_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(TEST_ANNOTE, "test_create_prt_prodigal-not-divisible3.lst")
    assert not prodigalfunc.create_prt(protfile, res_prt_file, exp_lst)
    assert ("Gene test.0417.00002.0001b_00003 has a number of nucleotides (3001) "
            "that is not divisible by 3.") in caplog.text


def test_create_prt_not_moreprots_lst(caplog):
    """
    Check that prt file is not generated when there are more proteins in lst than in faa
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "original_name.fna-prodigalRes",
                        "prodigal.outtest.ok.faa")
    res_prt_file = os.path.join(GENEPATH, "prodigal_res.gff")
    exp_lst = os.path.join(TEST_ANNOTE, "test_create_prt_prodigal-more-proteins.lst")
    assert not prodigalfunc.create_prt(protfile, res_prt_file, exp_lst)
    assert ("Protein test.0417.00002.0007b_00016 is in .lst file but its sequence is not "
            "in the protein file generated by prodigal.") in caplog.text


def test_format_1genome(caplog):
    """
    Test that when prodigal results are ok, all files are generated as expected.
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    name = "prodigal.outtest.ok"
    gpath =  os.path.join(TEST_ANNOTE, "original_name.fna") # path to original genome, given to prodigal for annotation
    prod_path = TEST_ANNOTE
    # Generate result folders
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gene_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(prot_dir)
    os.makedirs(lst_dir)
    os.makedirs(rep_dir)
    os.makedirs(gene_dir)
    os.makedirs(gff_dir)

    assert prodigalfunc.format_one_genome(gpath, name, prod_path, lst_dir, prot_dir, gene_dir,
                                          rep_dir, gff_dir)
    assert os.path.isfile(os.path.join(prot_dir, "prodigal.outtest.ok.prt"))
    assert os.path.isfile(os.path.join(lst_dir, "prodigal.outtest.ok.lst"))
    assert os.path.isfile(os.path.join(rep_dir, "prodigal.outtest.ok.fna"))
    assert os.path.isfile(os.path.join(gene_dir, "prodigal.outtest.ok.gen"))
    assert os.path.isfile(os.path.join(gff_dir, "prodigal.outtest.ok.gff"))


def test_format_1genome_emptygpath(caplog):
    """
    Test on formatting prodigal results, when ffn file is empty -> error message,
    and no file generated
    """
    caplog.set_level(logging.DEBUG)
    name = "prodigal.outtest.ok"
    # Create empty file, that we give to prodigal for formatting step
    gpath =  os.path.join(GENEPATH, "original_name-empty.fna")
    open(gpath, "w").close()
    prod_path = GENEPATH
    # Create prodigal result files (empty, then won't be read)
    prodigal_dir = gpath + "-prodigalRes"
    os.makedirs(prodigal_dir)
    prodigal_faa = os.path.join(gpath + "-prodigalRes", "notread.faa")
    prodigal_ffn = os.path.join(gpath + "-prodigalRes", "notread.ffn")
    prodigal_gff = os.path.join(gpath + "-prodigalRes", "notread.gff")
    for file in [prodigal_faa, prodigal_gff, prodigal_ffn]:
        open(file, "w").close()
    # Generate result folders
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gen_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(rep_dir)
    os.makedirs(gff_dir)
    os.makedirs(lst_dir)
    os.makedirs(gen_dir)
    # Add empty res lst, gff and gen files, to check that it is removed at the end
    res_gff_file = os.path.join(gff_dir, "prodigal.outtest.ok.gff")
    open(res_gff_file, "w").close()
    assert len(os.listdir(gff_dir) ) == 1
    res_lst_file = os.path.join(lst_dir, "prodigal.outtest.ok.lst")
    open(res_lst_file, "w").close()
    assert len(os.listdir(lst_dir) ) == 1
    res_gen_file = os.path.join(gen_dir, "prodigal.outtest.ok.gen")
    open(res_gen_file, "w").close()
    assert len(os.listdir(gen_dir) ) == 1

    assert not prodigalfunc.format_one_genome(gpath, name, prod_path, lst_dir, prot_dir, gen_dir,
                                              rep_dir, gff_dir)
    # Check that all files were removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert len(os.listdir(gen_dir) ) == 0
    assert ("Your genome test/data/annotate/generated_by_unit-tests/original_name-empty.fna "
            "does not contain any sequence, or is not in fasta format.") in caplog.text
    assert ("Problems while generating Replicon file for prodigal.outtest.ok") in caplog.text


def test_format_1genome_wrongffn(caplog):
    """
    Test on formatting prodigal results, when the ffn file generated by prodigal does
    not have the same contig name as in original fna file
    -> error message,
    -> no file generated
    """
    caplog.set_level(logging.DEBUG)
    name = "prodigal.outtest.ok"
    # path to original genome, given to prodigal for annotation
    orig_gpath =  os.path.join(TEST_ANNOTE, "original_name.fna")
    orig_prodpath = orig_gpath + "-prodigalRes"
    used_gpath = os.path.join(GENEPATH, "original_name.fna")
    used_respath = used_gpath + "-prodigalRes"
    # Add original genome, and prodigal results to result folder
    shutil.copyfile(orig_gpath, used_gpath)
    shutil.copytree(orig_prodpath, used_respath)
    # In GENEPATH folder, create the original genome given to prodigal
    # (copy from test_file)
    # Create gen_file with a header not existing
    with open(os.path.join(used_respath, "prodigal.outtest.ok.ffn"), "w") as ori:
        ori.write(">wrongheader # 1 # 2 # 1 # toto")
    prod_path = GENEPATH
    # Generate result folders
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gene_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(rep_dir)
    os.makedirs(gene_dir)
    os.makedirs(lst_dir)
    os.makedirs(gff_dir)
    os.makedirs(prot_dir)
    # Add empty res gff file, to check that it is removed at the end
    res_gff_file = os.path.join(gff_dir, "prodigal.outtest.ok.gff")
    open(res_gff_file, "w").close()
    assert len(os.listdir(gff_dir) ) == 1
    # Run formatting
    assert not prodigalfunc.format_one_genome(used_gpath, name, prod_path, lst_dir, prot_dir,
                                              gene_dir, rep_dir, gff_dir)
    # Check that all files were removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(gene_dir) ) == 0
    assert len(os.listdir(prot_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert ("'wrongheader' found in test/data/annotate/generated_by_unit-tests/"
            "original_name.fna-prodigalRes/prodigal.outtest.ok.ffn does not exist in "
            "test/data/annotate/generated_by_unit-tests/original_name.fna") in caplog.text
    assert ("Problems while generating .gen and .lst files for prodigal.outtest.ok") in caplog.text


def test_format_1genome_wronggff(caplog):
    """
    Test on formatting prodigal results, when the gff file generated by prodigal
    does not have same info as ffn file generated by prodigal (= lst file generated by panacota)
    -> error message,
    -> no file generated
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    name = "prodigal.outtest.ok"

    # path to original genome, given to prodigal for annotation
    orig_gpath =  os.path.join(TEST_ANNOTE, "original_name.fna")
    orig_prodpath = orig_gpath + "-prodigalRes"
    # In generated_by_tests folder, create the original genome given to prodigal
    # (copy from test_file)
    used_gpath = os.path.join(GENEPATH, "original_name.fna")
    used_respath = used_gpath + "-prodigalRes"
    # Add original genome, and prodigal results to result folder
    shutil.copyfile(orig_gpath, used_gpath)
    shutil.copytree(orig_prodpath, used_respath)
    # Copy gff file, but modify to get wrong start position
    orig_gff = os.path.join(orig_gpath + "-prodigalRes", "prodigal.outtest.ok.gff")
    used_gff = os.path.join(used_respath, "prodigal.outtest.ok.gff")
    with open(orig_gff, "r") as orig, open(used_gff, "w") as gffw:
        for line in orig:
            if line.startswith("#"):
                gffw.write(line)
            else:
                fields_g = line.split("\t")
                (contig_name, source, type_g, start_g, end_g,
                 score, strand_g, phase, attributes) = fields_g
                gffw.write("\t".join([contig_name, source, type_g, "12", end_g, score, strand_g,
                                      phase, attributes]))
                break
    prod_path = GENEPATH
    # Generate result folders
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gene_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff3")
    os.makedirs(rep_dir)
    os.makedirs(gene_dir)
    os.makedirs(prot_dir)
    os.makedirs(lst_dir)
    os.makedirs(gff_dir)
    assert not prodigalfunc.format_one_genome(used_gpath, name, prod_path, lst_dir, prot_dir,
                                              gene_dir, rep_dir, gff_dir)
    # Check that replicon and gff files were removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(gene_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert len(os.listdir(prot_dir) ) == 0
    assert ("Files prodigal.outtest.ok.ffn and prodigal.outtest.ok.gff (in prodigal tmp_files: "
            "test/data/annotate/generated_by_unit-tests/original_name.fna-prodigalRes) do "
            "not have the same start value for gene EPKOMDHM_00001 "
            "(12 in gff, 287 in ffn)") in caplog.text
    assert ("Problems while generating .gff (gff3 folder) file for "
            "prodigal.outtest.ok") in caplog.text



def test_format_1genome_wrongprt(caplog):
    """
    Test on formatting prodigal results, when the faa file generated by prodigal does
    not have as many proteins as the ffn file generated by prodigal (=lst file generated
    by panacota)
    -> error message,
    -> no file generated
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    name = "prodigal.outtest.ok"

    # path to original genome, given to prodigal for annotation
    orig_gpath =  os.path.join(TEST_ANNOTE, "original_name.fna")
    orig_prodpath = orig_gpath + "-prodigalRes"
    # In generated_by_tests folder, create the original genome given to prodigal
    # (copy from test_file)
    used_gpath = os.path.join(GENEPATH, "original_name.fna")
    used_respath = used_gpath + "-prodigalRes"
    # Add original genome, and prodigal results to result folder
    shutil.copyfile(orig_gpath, used_gpath)
    shutil.copytree(orig_prodpath, used_respath)
    orig_faa = os.path.join(orig_gpath + "-prodigalRes", "prodigal.outtest.ok.faa")
    used_faa = os.path.join(used_respath, "prodigal.outtest.ok.faa")
    with open(orig_faa, "r") as faa, open(used_faa, "w") as faar:
        for i in range(7):
            faa.readline()
        for line in faa:
            faar.write(line)
    # Generate result folders
    prod_path = GENEPATH
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gene_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff3")
    os.makedirs(rep_dir)
    os.makedirs(gene_dir)
    os.makedirs(lst_dir)
    os.makedirs(gff_dir)
    os.makedirs(prot_dir)
    # Copy generated lstfile, but modify first line to have a difference between gff and lst starts

    assert not prodigalfunc.format_one_genome(used_gpath, name, prod_path, lst_dir, prot_dir,
                                              gene_dir, rep_dir, gff_dir)
    # Check that replicon file was removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(gene_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert len(os.listdir(prot_dir) ) == 0
    assert ("Protein prodigal.outtest.ok.0007b_00013 is in .lst file but its sequence is not in "
            "the protein file generated by prodigal.") in caplog.text
    assert ("Problems while generating .prt file (Proteins folder) for "
            "prodigal.outtest.ok") in caplog.text
